#! /usr/bin/env python
from __future__ import print_function

import os
import subprocess
import sys

import pytest

PYTEST_ARGS = {
    'default': ['tests'],
    'fast': ['tests', '-q'],
}

FLAKE8_ARGS = ['drf_spectacular', 'tests']

MYPY_ARGS = ['--config-file=tox.ini', 'drf_spectacular', 'tests']

ISORT_ARGS = ['--check', '--diff', '.']

sys.path.append(os.path.dirname(__file__))


def exit_on_failure(ret, message=None):
    if ret:
        sys.exit(ret)


def flake8_main(args):
    print('Running flake8 code linting')
    ret = subprocess.call(['flake8'] + args)
    print('flake8 failed' if ret else 'flake8 passed')
    return ret


def mypy_main(args):
    print('Running mypy code linting')
    ret = subprocess.call(['mypy'] + args)
    print('mypy failed' if ret else 'mypy passed')
    return ret


def isort_main(args):
    print('Running isort code linting')
    ret = subprocess.call(['isort'] + args)
    print('isort failed, run: isort --interactive .' if ret else 'isort passed')
    return ret


def split_class_and_function(string):
    class_string, function_string = string.split('.', 1)
    return "%s and %s" % (class_string, function_string)


def is_function(string):
    # `True` if it looks like a test function is included in the string.
    return string.startswith('test_') or '.test_' in string


def is_class(string):
    # `True` if first character is uppercase - assume it's a class name.
    return string[0] == string[0].upper()


if __name__ == "__main__":
    try:
        sys.argv.remove('--nolint')
    except ValueError:
        run_lint = True
    else:
        run_lint = False

    try:
        sys.argv.remove('--lintonly')
    except ValueError:
        run_tests = True
    else:
        run_tests = False

    try:
        sys.argv.remove('--fast')
    except ValueError:
        style = 'default'
    else:
        style = 'fast'
        run_lint = False

    if len(sys.argv) > 1:
        pytest_args = sys.argv[1:]
        first_arg = pytest_args[0]
        if first_arg.startswith('-'):
            # `runtests.py [flags]`
            pytest_args = ['tests'] + pytest_args
        elif is_class(first_arg) and is_function(first_arg):
            # `runtests.py TestCase.test_function [flags]`
            expression = split_class_and_function(first_arg)
            pytest_args = ['tests', '-k', expression] + pytest_args[1:]
        elif is_class(first_arg) or is_function(first_arg):
            # `runtests.py TestCase [flags]`
            # `runtests.py test_function [flags]`
            pytest_args = ['tests', '-k', pytest_args[0]] + pytest_args[1:]
    else:
        pytest_args = PYTEST_ARGS[style]

    if run_tests:
        exit_on_failure(pytest.main(pytest_args))
    if run_lint:
        exit_on_failure(flake8_main(FLAKE8_ARGS))
        exit_on_failure(isort_main(ISORT_ARGS))
        exit_on_failure(mypy_main(MYPY_ARGS))



from django.apps import AppConfig


class SpectacularConfig(AppConfig):
    name = 'drf_spectacular'
    verbose_name = "drf-spectacular"

    def ready(self):
        import drf_spectacular.checks  # noqa: F401



from drf_spectacular.contrib.rest_polymorphic import PolymorphicSerializerExtension
from drf_spectacular.plumbing import ResolvedComponent
from drf_spectacular.serializers import PolymorphicProxySerializerExtension
from drf_spectacular.settings import spectacular_settings


class RollupMixin:
    """
    This is a schema helper that pulls the "common denominator" fields from child
    components into their parent component. It only applies to PolymorphicSerializer
    as well as PolymorphicProxySerializer, where there is an (implicit) inheritance hierarchy.

    The actual functionality is realized via extensions defined below.
    """
    def map_serializer(self, auto_schema, direction):
        schema = super().map_serializer(auto_schema, direction)

        if isinstance(self, PolymorphicProxySerializerExtension):
            sub_serializers = self.target.serializers
        else:
            sub_serializers = [
                self.target._get_serializer_from_model_or_instance(sub_model)
                for sub_model in self.target.model_serializer_mapping
            ]

        resolved_sub_serializers = [
            auto_schema.resolve_serializer(sub, direction) for sub in sub_serializers
        ]
        # this will only be generated on return of map_serializer so mock it for now
        mocked_component = ResolvedComponent(
            name=auto_schema._get_serializer_name(self.target, direction),
            type=ResolvedComponent.SCHEMA,
            object=self.target,
            schema=schema
        )

        # hack for recursive models. at the time of extension execution, not all sub
        # serializer schema have been generated, so no rollup is possible.
        # by registering a local variable scoped postproc hook, we delay this
        # execution to the end where all schemas are present.
        def postprocessing_rollup_hook(generator, result, **kwargs):
            rollup_properties(mocked_component, resolved_sub_serializers)
            result['components'] = generator.registry.build({})
            return result

        # register postproc hook. must run before enum postproc due to rebuilding the registry
        spectacular_settings.POSTPROCESSING_HOOKS.insert(0, postprocessing_rollup_hook)
        # and do nothing for now
        return schema


def rollup_properties(component, resolved_sub_serializers):
    # rollup already happened (spectacular bug and normally not needed)
    if any('allOf' in r.schema for r in resolved_sub_serializers):
        return

    all_field_sets = [
        set(list(r.schema['properties'])) for r in resolved_sub_serializers
    ]
    common_fields = all_field_sets[0].intersection(*all_field_sets[1:])
    common_schema = {
        'properties': {},
        'required': set(),
    }

    # substitute sub serializers' common fields with base class
    for r in resolved_sub_serializers:
        for cf in sorted(common_fields):
            if cf in r.schema['properties']:
                common_schema['properties'][cf] = r.schema['properties'][cf]
                del r.schema['properties'][cf]
                if cf in r.schema.get('required', []):
                    common_schema['required'].add(cf)
        r.schema = {'allOf': [component.ref, r.schema]}

    # modify regular schema for field rollup
    del component.schema['oneOf']
    component.schema['properties'] = common_schema['properties']
    if common_schema['required']:
        component.schema['required'] = sorted(common_schema['required'])


class PolymorphicRollupSerializerExtension(RollupMixin, PolymorphicSerializerExtension):
    priority = 1


class PolymorphicProxyRollupSerializerExtension(RollupMixin, PolymorphicProxySerializerExtension):
    priority = 1



from textwrap import dedent

from django.core.management.base import BaseCommand, CommandError
from django.utils import translation
from django.utils.module_loading import import_string

from drf_spectacular.drainage import GENERATOR_STATS
from drf_spectacular.renderers import OpenApiJsonRenderer, OpenApiYamlRenderer
from drf_spectacular.settings import patched_settings, spectacular_settings
from drf_spectacular.validation import validate_schema


class SchemaGenerationError(CommandError):
    pass


class SchemaValidationError(CommandError):
    pass


class Command(BaseCommand):
    help = dedent("""
        Generate a spectacular OpenAPI3-compliant schema for your API.

        The warnings serve as a indicator for where your API could not be properly
        resolved. @extend_schema and @extend_schema_field are your friends.
        The spec should be valid in any case. If not, please open an issue
        on github: https://github.com/tfranzel/drf-spectacular/issues

        Remember to configure your APIs meta data like servers, version, url,
        documentation and so on in your SPECTACULAR_SETTINGS."
    """)

    def add_arguments(self, parser):
        parser.add_argument('--format', dest="format", choices=['openapi', 'openapi-json'], default='openapi', type=str)
        parser.add_argument('--urlconf', dest="urlconf", default=None, type=str)
        parser.add_argument('--generator-class', dest="generator_class", default=None, type=str)
        parser.add_argument('--file', dest="file", default=None, type=str)
        parser.add_argument('--fail-on-warn', dest="fail_on_warn", default=False, action='store_true')
        parser.add_argument('--validate', dest="validate", default=False, action='store_true')
        parser.add_argument('--api-version', dest="api_version", default=None, type=str)
        parser.add_argument('--lang', dest="lang", default=None, type=str)
        parser.add_argument('--color', dest="color", default=False, action='store_true')
        parser.add_argument('--custom-settings', dest="custom_settings", default=None, type=str)

    def handle(self, *args, **options):
        if options['generator_class']:
            generator_class = import_string(options['generator_class'])
        else:
            generator_class = spectacular_settings.DEFAULT_GENERATOR_CLASS

        GENERATOR_STATS.enable_trace_lineno()

        if options['color']:
            GENERATOR_STATS.enable_color()

        generator = generator_class(
            urlconf=options['urlconf'],
            api_version=options['api_version'],
        )

        if options['custom_settings']:
            custom_settings = import_string(options['custom_settings'])
        else:
            custom_settings = None

        with patched_settings(custom_settings):
            if options['lang']:
                with translation.override(options['lang']):
                    schema = generator.get_schema(request=None, public=True)
            else:
                schema = generator.get_schema(request=None, public=True)

        GENERATOR_STATS.emit_summary()

        if options['fail_on_warn'] and GENERATOR_STATS:
            raise SchemaGenerationError('Failing as requested due to warnings')
        if options['validate']:
            try:
                validate_schema(schema)
            except Exception as e:
                raise SchemaValidationError(e)

        renderer = self.get_renderer(options['format'])
        output = renderer.render(schema, renderer_context={})

        if options['file']:
            with open(options['file'], 'wb') as f:
                f.write(output)
        else:
            self.stdout.write(output.decode())

    def get_renderer(self, format):
        renderer_cls = {
            'openapi': OpenApiYamlRenderer,
            'openapi-json': OpenApiJsonRenderer,
        }[format]
        return renderer_cls()



import pytest
from django.urls import path
from rest_framework import serializers
from rest_framework.decorators import api_view

from drf_spectacular.utils import extend_schema
from tests import assert_schema, generate_schema

try:
    from rest_framework_recursive.fields import RecursiveField
except ImportError:
    from rest_framework.fields import Field as RecursiveField


class TreeSerializer(serializers.Serializer):
    name = serializers.CharField()
    children = serializers.ListField(child=RecursiveField())


class TreeManySerializer(serializers.Serializer):
    name = serializers.CharField()
    children = RecursiveField(many=True)


class PingSerializer(serializers.Serializer):
    ping_id = serializers.IntegerField()
    pong = RecursiveField('PongSerializer', required=False)


class PongSerializer(serializers.Serializer):
    pong_id = serializers.IntegerField()
    ping = PingSerializer()


class LinkSerializer(serializers.Serializer):
    name = serializers.CharField(max_length=25)
    next = RecursiveField(allow_null=True)


@pytest.mark.contrib('rest_framework_recursive')
def test_rest_framework_recursive(no_warnings):
    @extend_schema(request=TreeSerializer, responses=TreeSerializer)
    @api_view(['POST'])
    def tree(request):
        pass  # pragma: no cover

    @extend_schema(request=TreeManySerializer, responses=TreeManySerializer)
    @api_view(['POST'])
    def tree_many(request):
        pass  # pragma: no cover

    @extend_schema(request=PingSerializer, responses=PingSerializer)
    @api_view(['POST'])
    def pong(request):
        pass  # pragma: no cover

    @extend_schema(request=LinkSerializer, responses=LinkSerializer)
    @api_view(['POST'])
    def link(request):
        pass  # pragma: no cover

    urlpatterns = [
        path('tree', tree),
        path('tree_many', tree_many),
        path('pong', pong),
        path('link', link)
    ]
    assert_schema(
        generate_schema(None, patterns=urlpatterns),
        'tests/contrib/test_rest_framework_recursive.yml'
    )



from drf_spectacular.drainage import error, warn
from drf_spectacular.extensions import OpenApiSerializerExtension
from drf_spectacular.plumbing import force_instance, is_list_serializer, is_serializer


class PolymorphicProxySerializerExtension(OpenApiSerializerExtension):
    target_class = 'drf_spectacular.utils.PolymorphicProxySerializer'
    priority = -1

    def get_name(self):
        return self.target.component_name

    def map_serializer(self, auto_schema, direction):
        """ custom handling for @extend_schema's injection of PolymorphicProxySerializer """
        if isinstance(self.target.serializers, dict):
            sub_components = self._get_explicit_sub_components(auto_schema, direction)
        else:
            sub_components = self._get_implicit_sub_components(auto_schema, direction)

        if not self._has_discriminator():
            return {'oneOf': [schema for _, schema in sub_components]}
        else:
            return {
                'oneOf': [schema for _, schema in sub_components],
                'discriminator': {
                    'propertyName': self.target.resource_type_field_name,
                    'mapping': {resource_type: schema['$ref'] for resource_type, schema in sub_components}
                }
            }

    def _get_implicit_sub_components(self, auto_schema, direction):
        sub_components = []
        for sub_serializer in self.target.serializers:
            sub_serializer = self._prep_serializer(sub_serializer)
            (resolved_name, resolved_schema) = self._process_serializer(auto_schema, sub_serializer, direction)
            if not resolved_schema:
                continue

            if not self._has_discriminator():
                sub_components.append((None, resolved_schema))
            else:
                try:
                    discriminator_field = sub_serializer.fields[self.target.resource_type_field_name]
                    resource_type = discriminator_field.to_representation(None)
                except:  # noqa: E722
                    warn(
                        f'sub-serializer {resolved_name} of {self.target.component_name} '
                        f'must contain the discriminator field "{self.target.resource_type_field_name}". '
                        f'defaulting to sub-serializer name, but schema will likely not match the API.'
                    )
                    resource_type = resolved_name
                sub_components.append((resource_type, resolved_schema))

        return sub_components

    def _get_explicit_sub_components(self, auto_schema, direction):
        sub_components = []
        for resource_type, sub_serializer in self.target.serializers.items():
            sub_serializer = self._prep_serializer(sub_serializer)
            (_, resolved_schema) = self._process_serializer(auto_schema, sub_serializer, direction)
            if resolved_schema:
                sub_components.append((resource_type, resolved_schema))

        return sub_components

    def _has_discriminator(self):
        return self.target.resource_type_field_name is not None

    def _prep_serializer(self, serializer):
        serializer = force_instance(serializer)
        serializer.partial = self.target.partial
        return serializer

    def _process_serializer(self, auto_schema, serializer, direction):
        if is_list_serializer(serializer):
            if self._has_discriminator() or self.target._many is not False:
                warn(
                    "To control sub-serializer's \"many\" attribute, following usage pattern is necessary: "
                    "PolymorphicProxySerializer(serializers=[...], resource_type_field_name=None, "
                    "many=False). Ignoring serializer as it is not processable in this configuration."
                )
                return None, None
            schema = auto_schema._unwrap_list_serializer(serializer, direction)
            return None, schema
        elif is_serializer(serializer):
            resolved = auto_schema.resolve_serializer(serializer, direction)
            return (resolved.name, resolved.ref) if resolved else (None, None)
        else:
            error("PolymorphicProxySerializer's serializer argument contained unknown objects.")
            return None, None



import difflib
import json
import os

from drf_spectacular.validation import validate_schema


def build_absolute_file_path(relative_path):
    return os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        relative_path
    )


def assert_schema(schema, reference_filename, transforms=None, reverse_transforms=None):
    from drf_spectacular.renderers import OpenApiJsonRenderer, OpenApiYamlRenderer

    schema_yml = OpenApiYamlRenderer().render(schema, renderer_context={})
    # render also a json and provoke serialization issues
    OpenApiJsonRenderer().render(schema, renderer_context={})

    reference_filename = build_absolute_file_path(reference_filename)

    with open(reference_filename.replace('.yml', '_out.yml'), 'wb') as fh:
        fh.write(schema_yml)

    if not os.path.exists(reference_filename):
        raise RuntimeError(
            f'{reference_filename} was not found for comparison. Carefully inspect '
            f'the generated {reference_filename.replace(".yml", "_out.yml")} and '
            f'copy it to {reference_filename} to serve as new ground truth.'
        )

    generated = schema_yml.decode()

    with open(reference_filename) as fh:
        expected = fh.read()

    # apply optional transformations to generated result. this mainly serves to unify
    # discrepancies between Django, DRF and library versions.
    for t in transforms or []:
        generated = t(generated)
    for t in reverse_transforms or []:
        expected = t(expected)

    assert_equal(generated, expected)
    # this is more a less a sanity check as checked-in schemas should be valid anyhow
    validate_schema(schema)


def assert_equal(actual, expected):
    if not isinstance(actual, str):
        actual = json.dumps(actual, indent=4)
    if not isinstance(expected, str):
        expected = json.dumps(expected, indent=4)
    diff = difflib.unified_diff(
        expected.splitlines(True),
        actual.splitlines(True),
    )
    diff = ''.join(diff)
    assert actual == expected and not diff, diff


def generate_schema(route, viewset=None, view=None, view_function=None, patterns=None):
    from django.urls import path
    from rest_framework import routers
    from rest_framework.viewsets import ViewSetMixin

    from drf_spectacular.generators import SchemaGenerator

    if viewset:
        assert issubclass(viewset, ViewSetMixin)
        router = routers.SimpleRouter()
        router.register(route, viewset, basename=route)
        patterns = router.urls
    elif view:
        patterns = [path(route, view.as_view())]
    elif view_function:
        patterns = [path(route, view_function)]
    else:
        assert route is None and isinstance(patterns, list)

    generator = SchemaGenerator(patterns=patterns)
    schema = generator.get_schema(request=None, public=True)
    validate_schema(schema)  # make sure generated schemas are always valid
    return schema


def get_response_schema(operation, status=None, content_type='application/json'):
    if (
        not status
        and operation['operationId'].endswith('_create')
        and '201' in operation['responses']
        and '200' not in operation['responses']
    ):
        status = '201'
    elif not status:
        status = '200'
    return operation['responses'][status]['content'][content_type]['schema']


def get_request_schema(operation, content_type='application/json'):
    return operation['requestBody']['content'][content_type]['schema']



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

from django.conf import settings

settings.configure(USE_I18N=False, USE_L10N=False)

sys.path[:0] = [os.path.abspath('../'), os.path.abspath('./')]

# -- Project information -----------------------------------------------------

project = 'drf-spectacular'
copyright = '2020, T. Franzel'
author = 'T. Franzel'

needs_sphinx = '4.1'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'extensions',
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

default_role = 'default-role-error'

linkcheck_allowed_redirects = {
    r"^https://tox\.wiki/$": r"https://tox\.wiki/en/latest/$",
    r"^https://drf-spectacular\.readthedocs\.io/$": r"https://drf-spectacular\.readthedocs\.io/en/latest/$",
    r"^https://docs\.djangoproject\.com/en/stable/": r"^https://docs\.djangoproject\.com/en/\d+\.\d+/",
    r"^https://github\.com/tfranzel/drf-spectacular/issues/\d+": "https://github\.com/tfranzel/drf-spectacular/pull/\d+",
}

linkcheck_ignore = [
    # Special-use addresses and domain names. (RFC 6761/6890)
    r"^https?://(?:127\.0\.0\.1|\[::1\])(?::\d+)?/",
    r"^https?://(?:[^/\.]+\.)*example\.(?:com|net|org)(?::\d+)?/",
    r"^https?://(?:[^/\.]+\.)*(?:example|invalid|localhost|test)(?::\d+)?/",
]

nitpicky = True

nitpick_ignore_regex = [
    # Unresolvable type hinting forward references.
    ('py:class', r'(?:APIView|AutoSchema|OpenApiFilterExtension)'),
    # Unresolvable type hinting references to packages without intersphinx support.
    ('py:class', r'rest_framework\..+'),
    # Internal undocumented objects.
    ('py:class', r'drf_spectacular\.generators\..+'),
    ('py:class', r'drf_spectacular\.plumbing\..+'),
    ('py:class', r'drf_spectacular\.utils\.F'),
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'django': ('https://docs.djangoproject.com/en/stable/', 'https://docs.djangoproject.com/en/stable/_objects/'),
    'drf-yasg': ('https://drf-yasg.readthedocs.io/en/stable/', None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'alabaster'
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



from drf_spectacular.extensions import OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import build_basic_type
from drf_spectacular.types import OpenApiTypes


class Base64FileFieldSchema(OpenApiSerializerFieldExtension):
    target_class = "drf_extra_fields.fields.Base64FileField"

    def map_serializer_field(self, auto_schema, direction):
        if direction == "request":
            return build_basic_type(OpenApiTypes.BYTE)
        elif direction == "response":
            if self.target.represent_in_base64:
                return build_basic_type(OpenApiTypes.BYTE)
            else:
                return build_basic_type(OpenApiTypes.URI)


class Base64ImageFieldSchema(Base64FileFieldSchema):
    target_class = "drf_extra_fields.fields.Base64ImageField"



from drf_spectacular.extensions import OpenApiViewExtension


class ObtainAuthTokenView(OpenApiViewExtension):
    target_class = 'rest_framework.authtoken.views.ObtainAuthToken'
    match_subclasses = True

    def view_replacement(self):
        """
        Prior to DRF 3.12.0, usage of ObtainAuthToken resulted in AssertionError

            Incompatible AutoSchema used on View "ObtainAuthToken". Is DRF's DEFAULT_SCHEMA_CLASS ...

        This is because DRF had a bug which made it NOT honor DEFAULT_SCHEMA_CLASS and instead
        injected an unsolicited coreschema class for this view and this view only. This extension
        fixes the view before the wrong schema class is used.

        Bug in DRF that was fixed in later versions:
        https://github.com/encode/django-rest-framework/blob/4121b01b912668c049b26194a9a107c27a332429/rest_framework/authtoken/views.py#L16
        """
        from rest_framework import VERSION

        from drf_spectacular.openapi import AutoSchema

        # no intervention needed
        if VERSION >= '3.12':
            return self.target

        class FixedObtainAuthToken(self.target):
            schema = AutoSchema()

        return FixedObtainAuthToken



import re


def camelize_serializer_fields(result, generator, request, public):
    from djangorestframework_camel_case.util import camelize_re, underscore_to_camel

    def camelize_str(str):
        return re.sub(camelize_re, underscore_to_camel, str)

    def camelize_component(schema: dict):
        if schema.get('type') == 'object':
            if 'properties' in schema:
                schema['properties'] = {
                    camelize_str(field_name): camelize_component(field_schema)
                    for field_name, field_schema in schema['properties'].items()
                }
            if 'required' in schema:
                schema['required'] = [camelize_str(field) for field in schema['required']]
        elif schema.get('type') == 'array':
            camelize_component(schema['items'])
        return schema

    for (_, component_type), component in generator.registry._components.items():
        if component_type == 'schemas':
            camelize_component(component.schema)

    # inplace modification of components also affect result dict, so regeneration is not necessary
    return result



import inspect
import os
from unittest import mock

import pytest
from django.urls import path
from rest_framework.test import APIClient

from drf_spectacular.views import SpectacularAPIView, SpectacularRedocView, SpectacularSwaggerView

urlpatterns = [
    path('api/schema/', SpectacularAPIView.as_view(), name='schema'),
    path('api/schema/swagger-ui/', SpectacularSwaggerView.as_view(), name='swagger'),
    path('api/schema/redoc/', SpectacularRedocView.as_view(), name='redoc'),
]

BUNDLE_URL = "static/drf_spectacular_sidecar/swagger-ui-dist/swagger-ui-bundle.js"


@mock.patch('drf_spectacular.settings.spectacular_settings.SWAGGER_UI_DIST', 'SIDECAR')
@mock.patch('drf_spectacular.settings.spectacular_settings.SWAGGER_UI_FAVICON_HREF', 'SIDECAR')
@mock.patch('drf_spectacular.settings.spectacular_settings.REDOC_DIST', 'SIDECAR')
@pytest.mark.urls(__name__)
@pytest.mark.contrib('drf_spectacular_sidecar')
def test_sidecar_shortcut_urls_are_resolved(no_warnings):
    response = APIClient().get('/api/schema/swagger-ui/')
    assert b'"/' + BUNDLE_URL.encode() + b'"' in response.content
    assert b'"/static/drf_spectacular_sidecar/swagger-ui-dist/favicon-32x32.png"' in response.content
    response = APIClient().get('/api/schema/redoc/')
    assert b'"/static/drf_spectacular_sidecar/redoc/bundles/redoc.standalone.js"' in response.content


@pytest.mark.contrib('drf_spectacular_sidecar')
def test_sidecar_package_urls_matching(no_warnings):
    # poor man's test to make sure the sidecar package contents match with what
    # collectstatic is going to compile. cannot be tested directly.
    import drf_spectacular_sidecar  # type: ignore[import]
    module_root = os.path.dirname(inspect.getfile(drf_spectacular_sidecar))
    bundle_path = os.path.join(module_root, BUNDLE_URL)
    assert os.path.isfile(bundle_path)



import enum
import typing
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from ipaddress import IPv4Address, IPv6Address
from uuid import UUID

_KnownPythonTypes = typing.Type[typing.Union[
    str, float, bool, bytes, int, dict, UUID, Decimal, datetime, date, time,
    timedelta, IPv4Address, IPv6Address,
]]


class OpenApiTypes(enum.Enum):
    """
    Basic types known to the OpenAPI specification or at least common format extension of it.

    - Use ``BYTE`` for base64-encoded data wrapped in a string
    - Use ``BINARY`` for raw binary data
    - Use ``OBJECT`` for arbitrary free-form object (usually a :py:class:`dict`)
    """
    #: Converted to ``{"type": "number"}``.
    NUMBER = enum.auto()
    #: Converted to ``{"type": "number", "format": "float"}``.
    #: Equivalent to :py:class:`float`.
    FLOAT = enum.auto()
    #: Converted to ``{"type": "number", "format": "double"}``.
    DOUBLE = enum.auto()
    #: Converted to ``{"type": "boolean"}``.
    #: Equivalent to :py:class:`bool`.
    BOOL = enum.auto()
    #: Converted to ``{"type": "string"}``.
    #: Equivalent to :py:class:`str`.
    STR = enum.auto()
    #: Converted to ``{"type": "string", "format": "byte"}``.
    #: Use this for base64-encoded data wrapped in a string.
    BYTE = enum.auto()
    #: Converted to ``{"type": "string", "format": "binary"}``.
    #: Equivalent to :py:class:`bytes`.
    #: Use this for raw binary data.
    BINARY = enum.auto()
    #: Converted to ``{"type": "string", "format": "password"}``.
    PASSWORD = enum.auto()
    #: Converted to ``{"type": "integer"}``.
    #: Equivalent to :py:class:`int`.
    INT = enum.auto()
    #: Converted to ``{"type": "integer", "format": "int32"}``.
    INT32 = enum.auto()
    #: Converted to ``{"type": "integer", "format": "int64"}``.
    INT64 = enum.auto()
    #: Converted to ``{"type": "string", "format": "uuid"}``.
    #: Equivalent to :py:class:`~uuid.UUID`.
    UUID = enum.auto()
    #: Converted to ``{"type": "string", "format": "uri"}``.
    URI = enum.auto()
    #: Converted to ``{"type": "string", "format": "uri-reference"}``.
    URI_REF = enum.auto()
    #: Converted to ``{"type": "string", "format": "uri-template"}``.
    URI_TPL = enum.auto()
    #: Converted to ``{"type": "string", "format": "iri"}``.
    IRI = enum.auto()
    #: Converted to ``{"type": "string", "format": "iri-reference"}``.
    IRI_REF = enum.auto()
    #: Converted to ``{"type": "string", "format": "ipv4"}``.
    #: Equivalent to :py:class:`~ipaddress.IPv4Address`.
    IP4 = enum.auto()
    #: Converted to ``{"type": "string", "format": "ipv6"}``.
    #: Equivalent to :py:class:`~ipaddress.IPv6Address`.
    IP6 = enum.auto()
    #: Converted to ``{"type": "string", "format": "hostname"}``.
    HOSTNAME = enum.auto()
    #: Converted to ``{"type": "string", "format": "idn-hostname"}``.
    IDN_HOSTNAME = enum.auto()
    #: Converted to ``{"type": "number", "format": "double"}``.
    #: The same as :py:attr:`~drf_spectacular.types.OpenApiTypes.DOUBLE`.
    #: Equivalent to :py:class:`~decimal.Decimal`.
    DECIMAL = enum.auto()
    #: Converted to ``{"type": "string", "format": "date-time"}``.
    #: Equivalent to :py:class:`~datetime.datetime`.
    DATETIME = enum.auto()
    #: Converted to ``{"type": "string", "format": "date"}``.
    #: Equivalent to :py:class:`~datetime.date`.
    DATE = enum.auto()
    #: Converted to ``{"type": "string", "format": "time"}``.
    #: Equivalent to :py:class:`~datetime.time`.
    TIME = enum.auto()
    #: Converted to ``{"type": "string", "format": "duration"}``.
    #: Equivalent to :py:class:`~datetime.timedelta`.
    #: Expressed according to ISO 8601.
    DURATION = enum.auto()
    #: Converted to ``{"type": "string", "format": "email"}``.
    EMAIL = enum.auto()
    #: Converted to ``{"type": "string", "format": "idn-email"}``.
    IDN_EMAIL = enum.auto()
    #: Converted to ``{"type": "string", "format": "json-pointer"}``.
    JSON_PTR = enum.auto()
    #: Converted to ``{"type": "string", "format": "relative-json-pointer"}``.
    JSON_PTR_REL = enum.auto()
    #: Converted to ``{"type": "string", "format": "regex"}``.
    REGEX = enum.auto()
    #: Converted to ``{"type": "object", ...}``.
    #: Use this for arbitrary free-form objects (usually a :py:class:`dict`).
    #: The ``additionalProperties`` item is added depending on the ``GENERIC_ADDITIONAL_PROPERTIES`` setting.
    OBJECT = enum.auto()
    #: Equivalent to :py:data:`None`.
    #: This signals that the request or response is empty.
    NONE = enum.auto()
    #: Converted to ``{}`` which sets no type and format.
    #: Equivalent to :py:class:`typing.Any`.
    ANY = enum.auto()


# make a copy with dict() before modifying returned dict
OPENAPI_TYPE_MAPPING = {
    OpenApiTypes.NUMBER: {'type': 'number'},
    OpenApiTypes.FLOAT: {'type': 'number', 'format': 'float'},
    OpenApiTypes.DOUBLE: {'type': 'number', 'format': 'double'},
    OpenApiTypes.BOOL: {'type': 'boolean'},
    OpenApiTypes.STR: {'type': 'string'},
    OpenApiTypes.BYTE: {'type': 'string', 'format': 'byte'},
    OpenApiTypes.BINARY: {'type': 'string', 'format': 'binary'},
    OpenApiTypes.PASSWORD: {'type': 'string', 'format': 'password'},
    OpenApiTypes.INT: {'type': 'integer'},
    OpenApiTypes.INT32: {'type': 'integer', 'format': 'int32'},
    OpenApiTypes.INT64: {'type': 'integer', 'format': 'int64'},
    OpenApiTypes.UUID: {'type': 'string', 'format': 'uuid'},
    OpenApiTypes.URI: {'type': 'string', 'format': 'uri'},
    OpenApiTypes.URI_REF: {'type': 'string', 'format': 'uri-reference'},
    OpenApiTypes.URI_TPL: {'type': 'string', 'format': 'uri-template'},
    OpenApiTypes.IRI: {'type': 'string', 'format': 'iri'},
    OpenApiTypes.IRI_REF: {'type': 'string', 'format': 'iri-reference'},
    OpenApiTypes.IP4: {'type': 'string', 'format': 'ipv4'},
    OpenApiTypes.IP6: {'type': 'string', 'format': 'ipv6'},
    OpenApiTypes.HOSTNAME: {'type': 'string', 'format': 'hostname'},
    OpenApiTypes.IDN_HOSTNAME: {'type': 'string', 'format': 'idn-hostname'},
    OpenApiTypes.DECIMAL: {'type': 'number', 'format': 'double'},
    OpenApiTypes.DATETIME: {'type': 'string', 'format': 'date-time'},
    OpenApiTypes.DATE: {'type': 'string', 'format': 'date'},
    OpenApiTypes.TIME: {'type': 'string', 'format': 'time'},
    OpenApiTypes.DURATION: {'type': 'string', 'format': 'duration'},  # ISO 8601
    OpenApiTypes.EMAIL: {'type': 'string', 'format': 'email'},
    OpenApiTypes.IDN_EMAIL: {'type': 'string', 'format': 'idn-email'},
    OpenApiTypes.JSON_PTR: {'type': 'string', 'format': 'json-pointer'},
    OpenApiTypes.JSON_PTR_REL: {'type': 'string', 'format': 'relative-json-pointer'},
    OpenApiTypes.REGEX: {'type': 'string', 'format': 'regex'},
    OpenApiTypes.ANY: {},
    OpenApiTypes.NONE: None,
    # OpenApiTypes.OBJECT is inserted at runtime due to dependency on settings
}

PYTHON_TYPE_MAPPING = {
    str: OpenApiTypes.STR,
    float: OpenApiTypes.DOUBLE,
    bool: OpenApiTypes.BOOL,
    bytes: OpenApiTypes.BINARY,
    int: OpenApiTypes.INT,
    UUID: OpenApiTypes.UUID,
    Decimal: OpenApiTypes.DECIMAL,
    datetime: OpenApiTypes.DATETIME,
    date: OpenApiTypes.DATE,
    time: OpenApiTypes.TIME,
    timedelta: OpenApiTypes.DURATION,
    IPv4Address: OpenApiTypes.IP4,
    IPv6Address: OpenApiTypes.IP6,
    dict: OpenApiTypes.OBJECT,
    typing.Any: OpenApiTypes.ANY,
    None: OpenApiTypes.NONE,
}

DJANGO_PATH_CONVERTER_MAPPING = {
    'int': OpenApiTypes.INT,
    'path': OpenApiTypes.STR,
    'slug': OpenApiTypes.STR,
    'str': OpenApiTypes.STR,
    'uuid': OpenApiTypes.UUID,
    'drf_format_suffix': OpenApiTypes.STR,
}



import os
from importlib import import_module

import django
import pytest
from django.core import management


def pytest_configure(config):
    from django.conf import settings

    config.addinivalue_line(
        "markers", "contrib(name): mark required contrib package"
    )

    contrib_apps = [
        'dj_rest_auth',
        'dj_rest_auth.registration',
        'allauth',
        'allauth.account',
        'oauth2_provider',
        'django_filters',
        'rest_framework_gis',
        # this is not strictly required and when added django-polymorphic
        # currently breaks the whole Django/DRF upstream testing.
        # 'polymorphic',
        # 'rest_framework_jwt',
    ]

    base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DATABASES={'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': ':memory:'
        }},
        SITE_ID=1,
        SECRET_KEY='not very secret in tests',
        USE_I18N=True,
        USE_L10N=True,
        LANGUAGES=[
            ('de-de', 'German'),
            ('en-us', 'English'),
        ],
        LOCALE_PATHS=[
            base_dir + '/locale/'
        ],
        STATIC_URL='/static/',
        ROOT_URLCONF='tests.urls',
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'DIRS': [],
                'APP_DIRS': True,
                'OPTIONS': {
                    'context_processors': [
                        'django.template.context_processors.debug',
                    ],
                },
            },
        ],
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.middleware.locale.LocaleMiddleware',
        ),
        INSTALLED_APPS=(
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.sites',
            'django.contrib.messages',
            'django.contrib.staticfiles',
            'rest_framework',
            'rest_framework.authtoken',
            *[app for app in contrib_apps if module_available(app)],
            'drf_spectacular',
            'tests',
        ),
        PASSWORD_HASHERS=(
            'django.contrib.auth.hashers.SHA1PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2PasswordHasher',
            'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
            'django.contrib.auth.hashers.BCryptPasswordHasher',
            'django.contrib.auth.hashers.MD5PasswordHasher',
            'django.contrib.auth.hashers.CryptPasswordHasher',
        ),
        REST_FRAMEWORK={
            'DEFAULT_SCHEMA_CLASS': 'drf_spectacular.openapi.AutoSchema',
            'PAGE_SIZE': 10,
        },
        DEFAULT_AUTO_FIELD='django.db.models.AutoField',
        SILENCED_SYSTEM_CHECKS=[
            'rest_framework.W001', 'fields.E210', 'security.W001', 'security.W002',
            'security.W003', 'security.W009', 'security.W012'
        ],
    )

    django.setup()
    # For whatever reason this works locally without an issue.
    # on TravisCI content_type table is missing in the sqlite db as
    # if no migration ran, but then why does it work locally?!
    management.call_command('migrate')


def pytest_addoption(parser):
    parser.addoption(
        "--skip-missing-contrib",
        action="store_true",
        default=False,
        help="skip tests depending on missing contrib packages"
    )
    parser.addoption(
        "--allow-contrib-fail",
        action="store_true",
        default=False,
        help="run contrib tests but allow them to fail"
    )


def pytest_collection_modifyitems(config, items):
    skip_missing_contrib = pytest.mark.skip(reason="skip tests for missing contrib package")
    allow_contrib_fail = pytest.mark.xfail(reason="contrib test were allowed to fail")
    for item in items:
        for marker in item.own_markers:
            if marker.name == 'contrib' and config.getoption("--skip-missing-contrib"):
                if not all([module_available(module_str) for module_str in marker.args]):  # pragma: no cover
                    item.add_marker(skip_missing_contrib)
            if marker.name == 'contrib' and config.getoption("--allow-contrib-fail"):
                item.add_marker(allow_contrib_fail)


@pytest.fixture()
def no_warnings(capsys):
    """ make sure test emits no warnings """
    yield capsys
    captured = capsys.readouterr()
    assert not captured.out
    assert not captured.err


@pytest.fixture()
def warnings(capsys):
    """ make sure test emits no warnings """
    yield capsys
    captured = capsys.readouterr()
    assert captured.err


@pytest.fixture()
def clear_generator_settings():
    from drf_spectacular.drainage import GENERATOR_STATS
    yield
    GENERATOR_STATS._trace_lineno = False
    GENERATOR_STATS._red = GENERATOR_STATS._blue = ''
    GENERATOR_STATS._yellow = GENERATOR_STATS._clear = ''


@pytest.fixture()
def clear_caches():
    from drf_spectacular.plumbing import get_openapi_type_mapping, load_enum_name_overrides
    load_enum_name_overrides.cache_clear()
    get_openapi_type_mapping.cache_clear()
    yield
    load_enum_name_overrides.cache_clear()
    get_openapi_type_mapping.cache_clear()


def module_available(module_str):
    try:
        import_module(module_str)
    except ImportError:
        return False
    else:
        return True



import functools
import json
import sys
import tempfile
import uuid
from datetime import timedelta
from decimal import Decimal
from typing import Optional

import pytest
from django import __version__ as DJANGO_VERSION
from django.core.files.base import ContentFile
from django.core.files.storage import FileSystemStorage
from django.db import models
from django.urls import reverse
from django.utils.functional import cached_property
from rest_framework import serializers, viewsets
from rest_framework.routers import SimpleRouter
from rest_framework.test import APIClient

from drf_spectacular.generators import SchemaGenerator
from tests import assert_equal, assert_schema, build_absolute_file_path

if sys.version_info >= (3, 8):
    functools_cached_property = functools.cached_property
else:
    # We re-use Django's cached_property when it's not avaiable to
    # keep tests unified across Python versions.
    functools_cached_property = cached_property


fs = FileSystemStorage(location=tempfile.gettempdir())


class Aux(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    field_foreign = models.ForeignKey('Aux', null=True, on_delete=models.CASCADE)
    url = models.URLField(unique=True, help_text="URL identifier for Aux")


class AuxSerializer(serializers.ModelSerializer):
    """ description for aux object """
    class Meta:
        fields = '__all__'
        model = Aux


class SubObject:
    def __init__(self, instance):
        self._instance = instance

    @property
    def calculated(self) -> int:
        return self._instance.field_int

    @property
    def nested(self) -> 'SubObject':
        return self

    @property
    def model_instance(self) -> 'AllFields':
        return self._instance

    @property
    def optional_int(self) -> Optional[int]:
        return 1


class AllFields(models.Model):
    # basics
    field_int = models.IntegerField()
    field_float = models.FloatField()
    field_bool = models.BooleanField()
    field_char = models.CharField(max_length=100)
    field_text = models.TextField(verbose_name='a text field')
    # special
    field_slug = models.SlugField()
    field_email = models.EmailField()
    field_uuid = models.UUIDField()
    field_url = models.URLField()
    field_ip_generic = models.GenericIPAddressField(protocol='ipv6')
    field_decimal = models.DecimalField(max_digits=6, decimal_places=3)
    field_file = models.FileField(storage=fs)
    field_img = models.ImageField(storage=fs)
    field_date = models.DateField()
    field_datetime = models.DateTimeField()
    field_bigint = models.BigIntegerField()
    field_smallint = models.SmallIntegerField()
    field_posint = models.PositiveIntegerField()
    field_possmallint = models.PositiveSmallIntegerField()
    if DJANGO_VERSION > '3.1':
        field_nullbool = models.BooleanField(null=True)
    else:
        field_nullbool = models.NullBooleanField()  # type: ignore
    field_time = models.TimeField()
    field_duration = models.DurationField()
    field_binary = models.BinaryField()

    # relations
    field_foreign = models.ForeignKey(
        Aux, on_delete=models.CASCADE, help_text='main aux object', related_name='ff'
    )
    field_m2m = models.ManyToManyField(
        Aux, help_text='set of related aux objects', related_name='fm'
    )
    field_o2o = models.OneToOneField(
        Aux, on_delete=models.CASCADE, help_text='bound aux object', related_name='fo'
    )
    # overrides
    field_regex = models.CharField(max_length=50)
    field_bool_override = models.BooleanField()

    if DJANGO_VERSION >= '3.1':
        field_json = models.JSONField()
    else:
        @property
        def field_json(self):
            return {'A': 1, 'B': 2}

    @property
    def field_model_property_float(self) -> float:
        return 1.337

    @cached_property
    def field_model_cached_property_float(self) -> float:
        return 1.337

    @functools_cached_property
    def field_model_py_cached_property_float(self) -> float:
        return 1.337

    @property
    def field_list(self):
        return [1.1, 2.2, 3.3]

    @property
    def field_list_object(self):
        return self.field_m2m.all()

    def model_function_basic(self) -> bool:
        return True

    def model_function_model(self) -> Aux:
        return self.field_foreign

    @property
    def sub_object(self) -> SubObject:
        return SubObject(self)

    @cached_property
    def sub_object_cached(self) -> SubObject:
        return SubObject(self)

    @functools_cached_property
    def sub_object_py_cached(self) -> SubObject:
        return SubObject(self)

    @property
    def optional_sub_object(self) -> Optional[SubObject]:
        return SubObject(self)


class AllFieldsSerializer(serializers.ModelSerializer):
    field_decimal_uncoerced = serializers.DecimalField(
        source='field_decimal',
        max_digits=6,
        decimal_places=3,
        coerce_to_string=False
    )
    field_method_float = serializers.SerializerMethodField()

    def get_field_method_float(self, obj) -> float:
        return 1.3456

    field_method_object = serializers.SerializerMethodField()

    def get_field_method_object(self, obj) -> dict:
        return {'key': 'value'}

    field_regex = serializers.RegexField(r'^[a-zA-z0-9]{10}\-[a-z]', label='A regex field')

    field_hidden = serializers.HiddenField(default='')

    # composite fields
    field_list = serializers.ListField(
        child=serializers.FloatField(), min_length=3, max_length=100,
    )
    field_list_serializer = serializers.ListField(
        child=AuxSerializer(),
        source='field_list_object',
    )

    # extra related fields
    field_related_slug = serializers.SlugRelatedField(
        read_only=True, source='field_foreign', slug_field='url',
    )  # type: ignore
    field_related_slug_queryset = serializers.SlugRelatedField(
        source='field_foreign', slug_field='url', queryset=Aux.objects.all()
    )  # type: ignore
    field_related_slug_many = serializers.SlugRelatedField(
        many=True, read_only=True, source='field_m2m', slug_field='url',
    )  # type: ignore
    field_related_string = serializers.StringRelatedField(
        source='field_foreign'
    )  # type: ignore
    field_related_hyperlink = serializers.HyperlinkedRelatedField(
        read_only=True, source='field_foreign', view_name='aux-detail'
    )  # type: ignore
    field_identity_hyperlink = serializers.HyperlinkedIdentityField(
        read_only=True, view_name='allfields-detail'
    )

    # read only - model traversal
    field_read_only_nav_uuid = serializers.ReadOnlyField(source='field_foreign.id')
    field_read_only_nav_uuid_3steps = serializers.ReadOnlyField(
        source='field_foreign.field_foreign.field_foreign.id',
        allow_null=True,  # force field output even if traversal fails
    )
    field_read_only_model_function_basic = serializers.ReadOnlyField(source='model_function_basic')
    field_read_only_model_function_model = serializers.ReadOnlyField(source='model_function_model.id')

    # override default writable bool field with readonly
    field_bool_override = serializers.ReadOnlyField()

    field_model_property_float = serializers.ReadOnlyField()

    field_model_cached_property_float = serializers.ReadOnlyField()

    field_model_py_cached_property_float = serializers.ReadOnlyField()

    field_dict_int = serializers.DictField(
        child=serializers.IntegerField(),
        source='field_json',
    )

    # there is a JSON model field for django>=3.1 that would be placed automatically. for <=3.1 we
    # need to set the field explicitly. defined here for both cases to have consistent ordering.
    field_json = serializers.JSONField()

    # traversal of non-model types of complex object
    field_sub_object_calculated = serializers.ReadOnlyField(source='sub_object.calculated')
    field_sub_object_nested_calculated = serializers.ReadOnlyField(source='sub_object.nested.calculated')
    field_sub_object_model_int = serializers.ReadOnlyField(source='sub_object.model_instance.field_int')

    field_sub_object_cached_calculated = serializers.ReadOnlyField(source='sub_object_cached.calculated')
    field_sub_object_cached_nested_calculated = serializers.ReadOnlyField(source='sub_object_cached.nested.calculated')
    field_sub_object_cached_model_int = serializers.ReadOnlyField(source='sub_object_cached.model_instance.field_int')

    field_sub_object_py_cached_calculated = serializers.ReadOnlyField(source='sub_object_py_cached.calculated')
    field_sub_object_py_cached_nested_calculated = serializers.ReadOnlyField(
        source='sub_object_py_cached.nested.calculated',
    )
    field_sub_object_py_cached_model_int = serializers.ReadOnlyField(
        source='sub_object_py_cached.model_instance.field_int',
    )

    # typing.Optional
    field_optional_sub_object_calculated = serializers.ReadOnlyField(
        source='optional_sub_object.calculated',
        allow_null=True,
    )
    field_sub_object_optional_int = serializers.ReadOnlyField(
        source='sub_object.optional_int',
        allow_null=True,
    )

    class Meta:
        fields = '__all__'
        model = AllFields


class AllFieldsModelViewset(viewsets.ReadOnlyModelViewSet):
    serializer_class = AllFieldsSerializer
    queryset = AllFields.objects.all()

    def retrieve(self, request, *args, **kwargs):
        return super().retrieve(request, *args, **kwargs)


class AuxModelViewset(viewsets.ReadOnlyModelViewSet):
    serializer_class = AuxSerializer
    queryset = Aux.objects.all()


router = SimpleRouter()
router.register('allfields', AllFieldsModelViewset)
router.register('aux', AuxModelViewset)
urlpatterns = router.urls


@pytest.mark.urls(__name__)
def test_fields(no_warnings):
    assert_schema(
        SchemaGenerator().get_schema(request=None, public=True),
        'tests/test_fields.yml'
    )


@pytest.mark.urls(__name__)
@pytest.mark.django_db
def test_model_setup_is_valid():
    aux = Aux(
        id='0ac6930d-87f4-40e8-8242-10a3ed31a335',
        url='https://xkcd.com'
    )
    aux.save()

    m = AllFields(
        # basics
        field_int=1,
        field_float=1.25,
        field_bool=True,
        field_char='char',
        field_text='text',
        # special
        field_slug='all_fields',
        field_email='test@example.com',
        field_uuid='00000000-00000000-00000000-00000000',
        field_url='https://github.com/tfranzel/drf-spectacular',
        field_ip_generic='2001:db8::8a2e:370:7334',
        field_decimal=Decimal('-666.333'),
        field_file=None,
        field_img=None,  # TODO fill with data below
        field_date='2021-09-09',
        field_datetime='2021-09-09T10:15:26.049862',
        field_bigint=11111111111111,
        field_smallint=111111,
        field_posint=123,
        field_possmallint=1,
        field_nullbool=None,
        field_time='00:05:23.283',
        field_duration=timedelta(seconds=10),
        field_binary=b'\xce\x9c\xce\xb9\xce\xb1',
        # relations
        field_foreign=aux,
        field_o2o=aux,
        # overrides
        field_regex='12345asdfg-a',
        field_bool_override=True,
    )
    if DJANGO_VERSION >= '3.1':
        m.field_json = {'A': 1, 'B': 2}
    m.field_file.save('hello.txt', ContentFile("hello world"), save=True)
    m.save()
    m.field_m2m.add(aux)

    response = APIClient().get(reverse('allfields-detail', args=(m.pk,)))
    assert response.status_code == 200

    with open(build_absolute_file_path('tests/test_fields_response.json')) as fh:
        expected = json.load(fh)

    if DJANGO_VERSION < '3':
        expected['field_file'] = f'http://testserver/allfields/1/{m.field_file.name}'
    else:
        expected['field_file'] = f'http://testserver/{m.field_file.name}'

    assert_equal(json.loads(response.content), expected)



import uuid

import pytest
from django.db import models
from rest_framework import mixins, serializers, viewsets
from rest_framework.renderers import JSONRenderer

from tests import assert_schema, generate_schema


class TreeNode(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    label = models.TextField()

    parent = models.ForeignKey(
        'TreeNode',
        null=True,
        blank=True,
        related_name='children',
        on_delete=models.DO_NOTHING
    )


class TreeNodeSerializer(serializers.ModelSerializer):
    class Meta:
        fields = ['id', 'label', 'parent', 'children']
        model = TreeNode

    def get_fields(self):
        fields = super(TreeNodeSerializer, self).get_fields()
        fields['children'] = TreeNodeSerializer(many=True)
        return fields


class TreeNodeViewset(mixins.RetrieveModelMixin, viewsets.GenericViewSet):
    serializer_class = TreeNodeSerializer
    queryset = TreeNode.objects.none()


def test_recursion(no_warnings):
    assert_schema(
        generate_schema('nodes', TreeNodeViewset),
        'tests/test_recursion.yml'
    )


@pytest.mark.django_db
def test_model_setup_is_valid():
    root = TreeNode(label='root')
    root.save()
    leaf1 = TreeNode(label='leaf1', parent=root)
    leaf1.save()
    leaf2 = TreeNode(label='leaf2', parent=root)
    leaf2.save()

    JSONRenderer().render(
        TreeNodeSerializer(root).data,
        accepted_media_type='application/json; indent=4'
    ).decode()



from django.conf import settings

from drf_spectacular.extensions import OpenApiSerializerFieldExtension
from drf_spectacular.plumbing import build_object_type


class TranslationsFieldFix(OpenApiSerializerFieldExtension):
    target_class = 'parler_rest.fields.TranslatedFieldsField'

    def map_serializer_field(self, auto_schema, direction):
        # Obtain auto-generated sub-serializer from parler_rest
        # Contains the fields wrapped in parler.models.TranslatedFields()
        translation_serializer = self.target.serializer_class
        # resolve translation sub-serializer into reusable component.
        translation_component = auto_schema.resolve_serializer(
            translation_serializer, direction
        )
        # advertise each language provided in PARLER_LANGUAGES
        return build_object_type(
            properties={
                i['code']: translation_component.ref for i in settings.PARLER_LANGUAGES[None]
            }
        )


from drf_spectacular.extensions import OpenApiAuthenticationExtension


class DjangoOAuthToolkitScheme(OpenApiAuthenticationExtension):
    target_class = 'oauth2_provider.contrib.rest_framework.OAuth2Authentication'
    name = 'oauth2'

    def get_security_requirement(self, auto_schema):
        from oauth2_provider.contrib.rest_framework import (
            IsAuthenticatedOrTokenHasScope, TokenHasScope, TokenMatchesOASRequirements,
        )
        view = auto_schema.view
        request = view.request

        for permission in auto_schema.view.get_permissions():
            if isinstance(permission, TokenMatchesOASRequirements):
                alt_scopes = permission.get_required_alternate_scopes(request, view)
                alt_scopes = alt_scopes.get(auto_schema.method, [])
                return [{self.name: group} for group in alt_scopes]
            if isinstance(permission, IsAuthenticatedOrTokenHasScope):
                return {self.name: TokenHasScope().get_scopes(request, view)}
            if isinstance(permission, TokenHasScope):
                # catch-all for subclasses of TokenHasScope like TokenHasReadWriteScope
                return {self.name: permission.get_scopes(request, view)}

    def get_security_definition(self, auto_schema):
        from oauth2_provider.scopes import get_scopes_backend

        from drf_spectacular.settings import spectacular_settings

        flows = {}
        for flow_type in spectacular_settings.OAUTH2_FLOWS:
            flows[flow_type] = {}
            if flow_type in ('implicit', 'authorizationCode'):
                flows[flow_type]['authorizationUrl'] = spectacular_settings.OAUTH2_AUTHORIZATION_URL
            if flow_type in ('password', 'clientCredentials', 'authorizationCode'):
                flows[flow_type]['tokenUrl'] = spectacular_settings.OAUTH2_TOKEN_URL
            if spectacular_settings.OAUTH2_REFRESH_URL:
                flows[flow_type]['refreshUrl'] = spectacular_settings.OAUTH2_REFRESH_URL
            scope_backend = get_scopes_backend()
            flows[flow_type]['scopes'] = scope_backend.get_all_scopes()

        return {
            'type': 'oauth2',
            'flows': flows
        }



import os
import re

from django.urls import URLPattern, URLResolver
from rest_framework import views, viewsets
from rest_framework.schemas.generators import BaseSchemaGenerator  # type: ignore
from rest_framework.schemas.generators import EndpointEnumerator as BaseEndpointEnumerator
from rest_framework.settings import api_settings

from drf_spectacular.drainage import add_trace_message, get_override, reset_generator_stats
from drf_spectacular.extensions import OpenApiViewExtension
from drf_spectacular.openapi import AutoSchema
from drf_spectacular.plumbing import (
    ComponentRegistry, alpha_operation_sorter, build_root_object, camelize_operation, error,
    get_class, is_versioning_supported, modify_for_versioning, normalize_result_object,
    operation_matches_version, sanitize_result_object, warn,
)
from drf_spectacular.settings import spectacular_settings


class EndpointEnumerator(BaseEndpointEnumerator):
    def get_api_endpoints(self, patterns=None, prefix=''):
        api_endpoints = self._get_api_endpoints(patterns, prefix)

        for hook in spectacular_settings.PREPROCESSING_HOOKS:
            api_endpoints = hook(endpoints=api_endpoints)

        api_endpoints_deduplicated = {}
        for path, path_regex, method, callback in api_endpoints:
            if (path, method) not in api_endpoints_deduplicated:
                api_endpoints_deduplicated[path, method] = (path, path_regex, method, callback)

        api_endpoints = list(api_endpoints_deduplicated.values())

        if callable(spectacular_settings.SORT_OPERATIONS):
            return sorted(api_endpoints, key=spectacular_settings.SORT_OPERATIONS)
        elif spectacular_settings.SORT_OPERATIONS:
            return sorted(api_endpoints, key=alpha_operation_sorter)
        else:
            return api_endpoints

    def get_path_from_regex(self, path_regex):
        path = super().get_path_from_regex(path_regex)
        # bugfix oversight in DRF regex stripping
        path = path.replace('\\.', '.')
        return path

    def _get_api_endpoints(self, patterns, prefix):
        """
        Return a list of all available API endpoints by inspecting the URL conf.
        Only modification the the DRF version is passing through the path_regex.
        """
        if patterns is None:
            patterns = self.patterns

        api_endpoints = []

        for pattern in patterns:
            path_regex = prefix + str(pattern.pattern)
            if isinstance(pattern, URLPattern):
                path = self.get_path_from_regex(path_regex)
                callback = pattern.callback
                if self.should_include_endpoint(path, callback):
                    for method in self.get_allowed_methods(callback):
                        endpoint = (path, path_regex, method, callback)
                        api_endpoints.append(endpoint)

            elif isinstance(pattern, URLResolver):
                nested_endpoints = self._get_api_endpoints(
                    patterns=pattern.url_patterns,
                    prefix=path_regex
                )
                api_endpoints.extend(nested_endpoints)

        return api_endpoints

    def get_allowed_methods(self, callback):
        methods = super().get_allowed_methods(callback)
        return [
            method for method in methods
            if method not in ('OPTIONS', 'HEAD', 'TRACE', 'CONNECT')
        ]


class SchemaGenerator(BaseSchemaGenerator):
    endpoint_inspector_cls = EndpointEnumerator

    def __init__(self, *args, **kwargs):
        self.registry = ComponentRegistry()
        self.api_version = kwargs.pop('api_version', None)
        self.inspector = None
        super().__init__(*args, **kwargs)

    def coerce_path(self, path, method, view):
        """
        Customized coerce_path which also considers the `_pk` suffix in URL paths
        of nested routers.
        """
        path = super().coerce_path(path, method, view)  # take care of {pk}
        if spectacular_settings.SCHEMA_COERCE_PATH_PK_SUFFIX:
            path = re.sub(pattern=r'{(\w+)_pk}', repl=r'{\1_id}', string=path)
        return path

    def create_view(self, callback, method, request=None):
        """
        customized create_view which is called when all routes are traversed. part of this
        is instantiating views with default params. in case of custom routes (@action) the
        custom AutoSchema is injected properly through 'initkwargs' on view. However, when
        decorating plain views like retrieve, this initialization logic is not running.
        Therefore forcefully set the schema if @extend_schema decorator was used.
        """
        override_view = OpenApiViewExtension.get_match(callback.cls)
        if override_view:
            original_cls = callback.cls
            callback.cls = override_view.view_replacement()

        # we refrain from passing request and deal with it ourselves in parse()
        view = super().create_view(callback, method, None)

        # drf-yasg compatibility feature. makes the view aware that we are running
        # schema generation and not a real request.
        view.swagger_fake_view = True

        # callback.cls is hosted in urlpatterns and is therefore not an ephemeral modification.
        # restore after view creation so potential revisits have a clean state as basis.
        if override_view:
            callback.cls = original_cls

        if isinstance(view, viewsets.ViewSetMixin):
            action = getattr(view, view.action)
        elif isinstance(view, views.APIView):
            action = getattr(view, method.lower())
        else:
            error(
                'Using not supported View class. Class must be derived from APIView '
                'or any of its subclasses like GenericApiView, GenericViewSet.'
            )
            return view

        action_schema = getattr(action, 'kwargs', {}).get('schema', None)
        if not action_schema:
            # there is no method/action customized schema so we are done here.
            return view

        # action_schema is either a class or instance. when @extend_schema is used, it
        # is always a class to prevent the weakref reverse "schema.view" bug for multi
        # annotations. The bug is prevented by delaying the instantiation of the schema
        # class until create_view (here) and not doing it immediately in @extend_schema.
        action_schema_class = get_class(action_schema)
        view_schema_class = get_class(callback.cls.schema)

        if not issubclass(action_schema_class, view_schema_class):
            # this handles the case of having a manually set custom AutoSchema on the
            # view together with extend_schema. In most cases, the decorator mechanics
            # prevent extend_schema from having access to the view's schema class. So
            # extend_schema is forced to use DEFAULT_SCHEMA_CLASS as fallback base class
            # instead of the correct base class set in view. We remedy this chicken-egg
            # problem here by rearranging the class hierarchy.
            mro = tuple(
                cls for cls in action_schema_class.__mro__
                if cls not in api_settings.DEFAULT_SCHEMA_CLASS.__mro__
            ) + view_schema_class.__mro__
            action_schema_class = type('ExtendedRearrangedSchema', mro, {})

        view.schema = action_schema_class()
        return view

    def _initialise_endpoints(self):
        if self.endpoints is None:
            self.inspector = self.endpoint_inspector_cls(self.patterns, self.urlconf)
            self.endpoints = self.inspector.get_api_endpoints()

    def _get_paths_and_endpoints(self):
        """
        Generate (path, method, view) given (path, method, callback) for paths.
        """
        view_endpoints = []
        for path, path_regex, method, callback in self.endpoints:
            view = self.create_view(callback, method)
            path = self.coerce_path(path, method, view)
            view_endpoints.append((path, path_regex, method, view))

        return view_endpoints

    def parse(self, input_request, public):
        """ Iterate endpoints generating per method path operations. """
        result = {}
        self._initialise_endpoints()
        endpoints = self._get_paths_and_endpoints()

        if spectacular_settings.SCHEMA_PATH_PREFIX is None:
            # estimate common path prefix if none was given. only use it if we encountered more
            # than one view to prevent emission of erroneous and unnecessary fallback names.
            non_trivial_prefix = len(set([view.__class__ for _, _, _, view in endpoints])) > 1
            if non_trivial_prefix:
                path_prefix = os.path.commonpath([path for path, _, _, _ in endpoints])
                path_prefix = re.escape(path_prefix)  # guard for RE special chars in path
            else:
                path_prefix = '/'
        else:
            path_prefix = spectacular_settings.SCHEMA_PATH_PREFIX
        if not path_prefix.startswith('^'):
            path_prefix = '^' + path_prefix  # make sure regex only matches from the start

        for path, path_regex, method, view in endpoints:
            # emit queued up warnings/error that happened prior to generation (decoration)
            for w in get_override(view, 'warnings', []):
                warn(w)
            for e in get_override(view, 'errors', []):
                error(e)

            view.request = spectacular_settings.GET_MOCK_REQUEST(method, path, view, input_request)

            if not (public or self.has_view_permissions(path, method, view)):
                continue

            if view.versioning_class and not is_versioning_supported(view.versioning_class):
                warn(
                    f'using unsupported versioning class "{view.versioning_class}". view will be '
                    f'processed as unversioned view.'
                )
            elif view.versioning_class:
                version = (
                    self.api_version  # explicit version from CLI, SpecView or SpecView request
                    or view.versioning_class.default_version  # fallback
                )
                if not version:
                    continue
                path = modify_for_versioning(self.inspector.patterns, method, path, view, version)
                if not operation_matches_version(view, version):
                    continue

            assert isinstance(view.schema, AutoSchema), (
                f'Incompatible AutoSchema used on View {view.__class__}. Is DRF\'s '
                f'DEFAULT_SCHEMA_CLASS pointing to "drf_spectacular.openapi.AutoSchema" '
                f'or any other drf-spectacular compatible AutoSchema?'
            )
            with add_trace_message(getattr(view, '__class__', view)):
                operation = view.schema.get_operation(
                    path, path_regex, path_prefix, method, self.registry
                )

            # operation was manually removed via @extend_schema
            if not operation:
                continue

            if spectacular_settings.SCHEMA_PATH_PREFIX_TRIM:
                path = re.sub(pattern=path_prefix, repl='', string=path, flags=re.IGNORECASE)

            if spectacular_settings.SCHEMA_PATH_PREFIX_INSERT:
                path = spectacular_settings.SCHEMA_PATH_PREFIX_INSERT + path

            if not path.startswith('/'):
                path = '/' + path

            if spectacular_settings.CAMELIZE_NAMES:
                path, operation = camelize_operation(path, operation)

            result.setdefault(path, {})
            result[path][method.lower()] = operation

        return result

    def get_schema(self, request=None, public=False):
        """ Generate a OpenAPI schema. """
        reset_generator_stats()
        result = build_root_object(
            paths=self.parse(request, public),
            components=self.registry.build(spectacular_settings.APPEND_COMPONENTS),
            version=self.api_version or getattr(request, 'version', None),
        )
        for hook in spectacular_settings.POSTPROCESSING_HOOKS:
            result = hook(result=result, generator=self, request=request, public=public)

        return sanitize_result_object(normalize_result_object(result))



from drf_spectacular.drainage import get_override, has_override
from drf_spectacular.extensions import OpenApiSerializerExtension
from drf_spectacular.plumbing import get_doc
from drf_spectacular.utils import Direction


class OpenApiDataclassSerializerExtensions(OpenApiSerializerExtension):
    target_class = "rest_framework_dataclasses.serializers.DataclassSerializer"
    match_subclasses = True

    def get_name(self):
        """Use the dataclass name in the schema, instead of the serializer prefix (which can be just Dataclass)."""
        if getattr(getattr(self.target, 'Meta', None), 'ref_name', None) is not None:
            return self.target.Meta.ref_name
        if has_override(self.target.dataclass_definition.dataclass_type, 'component_name'):
            return get_override(self.target.dataclass_definition.dataclass_type, 'component_name')
        return self.target.dataclass_definition.dataclass_type.__name__

    def strip_library_doc(self, schema):
        """Strip the DataclassSerializer library documentation from the schema."""
        from rest_framework_dataclasses.serializers import DataclassSerializer
        if 'description' in schema and schema['description'] == get_doc(DataclassSerializer):
            del schema['description']
        return schema

    def map_serializer(self, auto_schema, direction: Direction):
        """"Generate the schema for a DataclassSerializer."""
        schema = auto_schema._map_serializer(self.target, direction, bypass_extensions=True)
        return self.strip_library_doc(schema)



from docutils.nodes import Text
from sphinx.util import logging


logger = logging.getLogger(__name__)


def setup(app):
    app.add_role("default-role-error", default_role_error)
    return {"parallel_read_safe": True}


# Inspired by equivalent code in Django's Sphinx extension.
# https://github.com/django/django/blob/6b53114d/docs/_ext/djangodocs.py#L393
def default_role_error(
    name, rawtext, text, lineno, inliner, options=None, content=None
):
    logger.warning(
        (
            f"Default role used (`single backticks`): {rawtext}. Did you mean to use "
            "two backticks for ``code``, or miss an underscore for a `link`_ ?"
        ),
        location=(inliner.document.current_source, lineno),
    )
    return [Text(text)], []



from djstripe.contrib.rest_framework.serializers import (
    CreateSubscriptionSerializer, SubscriptionSerializer
)

from drf_spectacular.extensions import OpenApiViewExtension
from drf_spectacular.utils import extend_schema


class FixDjstripeSubscriptionRestView(OpenApiViewExtension):
    target_class = 'djstripe.contrib.rest_framework.views.SubscriptionRestView'

    def view_replacement(self):
        class Fixed(self.target_class):
            serializer_class = SubscriptionSerializer

            @extend_schema(
                request=CreateSubscriptionSerializer,
                responses=CreateSubscriptionSerializer
            )
            def post(self, request, *args, **kwargs):
                pass

        return Fixed



import pytest
import yaml
from django.contrib.auth.models import User
from django.urls import include, path
from rest_framework import routers, viewsets
from rest_framework.authentication import TokenAuthentication
from rest_framework.authtoken.models import Token
from rest_framework.permissions import IsAuthenticated
from rest_framework.test import APIClient
from rest_framework.versioning import AcceptHeaderVersioning

from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.views import SpectacularAPIView
from tests.models import SimpleModel, SimpleSerializer


class AnotherSimpleSerializer(SimpleSerializer):
    pass


class XViewset(viewsets.ModelViewSet):
    authentication_classes = [TokenAuthentication]
    permission_classes = [IsAuthenticated]
    queryset = SimpleModel.objects.none()

    def get_serializer_class(self):
        # make sure the mocked request possesses the correct path and
        # schema endpoint path does not leak in.
        assert self.request.path.startswith('/api/x/')
        # make schema dependent on request method
        if self.request.method == 'GET':
            return SimpleSerializer
        else:
            return AnotherSimpleSerializer


router = routers.SimpleRouter()
router.register('x', XViewset)
urlpatterns = [
    path('api/', include(router.urls)),
    path('api/schema-plain/', SpectacularAPIView.as_view()),
    path('api/schema-authenticated/', SpectacularAPIView.as_view(
        authentication_classes=[TokenAuthentication]
    )),
    path('api/schema-authenticated-private/', SpectacularAPIView.as_view(
        authentication_classes=[TokenAuthentication],
        serve_public=False,
    )),
    path('api/schema-versioned/', SpectacularAPIView.as_view(
        versioning_class=AcceptHeaderVersioning
    ))
]


@pytest.mark.urls(__name__)
def test_mock_request_symmetry_plain(no_warnings):
    response = APIClient().get('/api/schema-plain/', **{'HTTP_X_SPECIAL_HEADER': '1'})
    assert response.status_code == 200
    schema_online = yaml.load(response.content, Loader=yaml.SafeLoader)
    schema_offline = SchemaGenerator().get_schema(public=True)
    assert schema_offline == schema_online


@pytest.mark.urls(__name__)
def test_mock_request_symmetry_version(no_warnings):
    response = APIClient().get('/api/schema-versioned/', **{
        'HTTP_ACCEPT': 'application/json; version=v2',
    })
    assert response.status_code == 200
    schema_online = yaml.load(response.content, Loader=yaml.SafeLoader)
    schema_offline = SchemaGenerator(api_version='v2').get_schema(public=True)

    assert schema_offline == schema_online
    assert schema_online['info']['version'] == '0.0.0 (v2)'


@pytest.mark.parametrize(['serve_public', 'authenticated', 'url', 'expected_endpoints'], [
    (True, True, '/api/schema-authenticated/', 5),
    (True, False, '/api/schema-authenticated/', 5),
    (False, True, '/api/schema-authenticated-private/', None),
    (False, False, '/api/schema-authenticated-private/', 3),
])
@pytest.mark.urls(__name__)
@pytest.mark.django_db
def test_mock_request_symmetry_authentication(
        no_warnings, serve_public, authenticated, url, expected_endpoints
):
    user = User.objects.create(username='test')
    token, _ = Token.objects.get_or_create(user=user)
    auth_header = {'HTTP_AUTHORIZATION': f'Token {token}'} if authenticated else {}
    response = APIClient().get(url, **auth_header)
    assert response.status_code == 200

    schema_online = yaml.load(response.content, Loader=yaml.SafeLoader)
    schema_offline = SchemaGenerator().get_schema(public=serve_public)

    if expected_endpoints:
        assert schema_offline == schema_online
        assert len(schema_online['paths']) == expected_endpoints
    else:
        # authenticated & non-public case does not really make sense for
        # offline generation as we have no request.
        assert len(schema_online['paths']) == 5
        assert len(schema_offline['paths']) == 3



from django.core.checks import Error, Warning, register


@register(deploy=True)
def schema_check(app_configs, **kwargs):
    """ Perform dummy generation and emit warnings/errors as part of Django's check framework """
    from drf_spectacular.drainage import GENERATOR_STATS
    from drf_spectacular.settings import spectacular_settings

    if not spectacular_settings.ENABLE_DJANGO_DEPLOY_CHECK:
        return []

    errors = []
    try:
        with GENERATOR_STATS.silence():
            spectacular_settings.DEFAULT_GENERATOR_CLASS().get_schema(request=None, public=True)
    except Exception as exc:
        errors.append(
            Error(f'Schema generation threw exception "{exc}"', id='drf_spectacular.E001')
        )
    if GENERATOR_STATS:
        for w in GENERATOR_STATS._warn_cache:
            errors.append(Warning(w, id='drf_spectacular.W001'))
        for e in GENERATOR_STATS._error_cache:
            errors.append(Warning(e, id='drf_spectacular.W002'))
    return errors



from rest_framework.renderers import TemplateHTMLRenderer
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView

from drf_spectacular.plumbing import get_relative_url, set_query_parameters
from drf_spectacular.settings import spectacular_settings
from drf_spectacular.utils import extend_schema
from drf_spectacular.views import AUTHENTICATION_CLASSES


class SpectacularElementsView(APIView):
     renderer_classes = [TemplateHTMLRenderer]
     permission_classes = spectacular_settings.SERVE_PERMISSIONS
     authentication_classes = AUTHENTICATION_CLASSES
     url_name = 'schema'
     url = None
     template_name = 'elements.html'
     title = spectacular_settings.TITLE

     @extend_schema(exclude=True)
     def get(self, request, *args, **kwargs):
        schema_url = self.url or get_relative_url(reverse(self.url_name, request=request))
        schema_url = set_query_parameters(schema_url, lang=request.GET.get('lang'), version=request.GET.get('version'))
        return Response(
            data={
                'title': self.title,
                'js_dist': 'https://unpkg.com/@stoplight/elements/web-components.min.js',
                'css_dist': 'https://unpkg.com/@stoplight/elements/styles.min.css',
                'schema_url': self._get_schema_url(request),
            },
            template_name=self.template_name
        )



# type: ignore
import re
from unittest import mock

import pytest
import yaml
from django.conf.urls import include
from django.db import models
from django.urls import path, re_path
from rest_framework import generics, mixins, routers, serializers, viewsets
from rest_framework.test import APIClient, APIRequestFactory
from rest_framework.versioning import AcceptHeaderVersioning, NamespaceVersioning, URLPathVersioning

from drf_spectacular.generators import SchemaGenerator
from drf_spectacular.utils import extend_schema
from drf_spectacular.validation import validate_schema
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView
from tests import assert_schema
from tests.models import SimpleModel


class Xv1Serializer(serializers.Serializer):
    id = serializers.IntegerField()


class Xv2Serializer(serializers.Serializer):
    id = serializers.UUIDField()


class PathVersioningViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
    versioning_class = URLPathVersioning
    queryset = SimpleModel.objects.all()

    @extend_schema(responses=Xv1Serializer, versions=['v1'])
    @extend_schema(responses=Xv2Serializer, versions=['v2'])
    def list(self, request, *args, **kwargs):
        pass  # pragma: no cover

    @extend_schema(responses=Xv1Serializer, versions=['v1'])
    @extend_schema(responses=Xv2Serializer, versions=['v2'])
    def retrieve(self, request, *args, **kwargs):
        pass  # pragma: no cover


class NamespaceVersioningViewset(PathVersioningViewset):
    versioning_class = NamespaceVersioning


class AcceptHeaderVersioningViewset(PathVersioningViewset):
    versioning_class = AcceptHeaderVersioning


class PathVersioningViewset2(mixins.ListModelMixin, viewsets.GenericViewSet):
    versioning_class = URLPathVersioning
    queryset = SimpleModel.objects.all()

    def list(self, request, *args, **kwargs):
        pass  # pragma: no cover

    def retrieve(self, request, *args, **kwargs):
        pass  # pragma: no cover

    def get_serializer_class(self):
        if self.request.version == 'v2':
            return Xv2Serializer
        return Xv1Serializer


class NamespaceVersioningViewset2(PathVersioningViewset2):
    versioning_class = NamespaceVersioning


class AcceptHeaderVersioningViewset2(PathVersioningViewset2):
    versioning_class = AcceptHeaderVersioning


@pytest.mark.parametrize('viewset_cls', [PathVersioningViewset, PathVersioningViewset2])
@pytest.mark.parametrize('version', ['v1', 'v2'])
def test_url_path_versioning(no_warnings, viewset_cls, version):
    router = routers.SimpleRouter()
    router.register('x', viewset_cls, basename='x')
    generator = SchemaGenerator(
        patterns=[re_path(r'^(?P<version>[v1|v2]+)/', include((router.urls, 'x')))],
        api_version=version,
    )
    schema = generator.get_schema(request=None, public=True)
    assert_schema(schema, f'tests/test_versioning_{version}.yml')


@pytest.mark.parametrize('viewset_cls', [NamespaceVersioningViewset, NamespaceVersioningViewset2])
@pytest.mark.parametrize('version', ['v1', 'v2'])
def test_namespace_versioning(no_warnings, viewset_cls, version):
    router = routers.SimpleRouter()
    router.register('x', viewset_cls, basename='x')
    generator = SchemaGenerator(
        patterns=[
            path('v1/', include((router.urls, 'v1'))),
            path('v2/', include((router.urls, 'v2'))),
        ],
        api_version=version,
    )
    schema = generator.get_schema(request=None, public=True)
    assert_schema(schema, f'tests/test_versioning_{version}.yml')


class LookupModel(models.Model):
    """ test_namespace_versioning_urlpatterns_simplification """
    field = models.IntegerField()


@pytest.mark.parametrize(['path_func', 'path_str', 'pattern', ], [
    (path, '{id}/', '<int:pk>/'),
    (path, '{id}/', '<pk>/'),
    (re_path, '{id}/', r'(?P<pk>[0-9A-Fa-f-]+)/'),
    (re_path, '{id}/', r'(?P<pk>[^/.]+)/$'),
    (re_path, '{id}/', r'(?P<pk>[a-z]{2}(-[a-z]{2})?)/'),
    (re_path, '{field}/t/{id}/', r'^(?P<field>[^/.]+)/t/(?P<pk>[^/.]+)/'),
    (re_path, '{field}/t/{id}/', r'^(?P<field>[A-Z\(\)]+)/t/(?P<pk>[^/.]+)/'),
])
def test_namespace_versioning_urlpatterns_simplification(no_warnings, path_func, path_str, pattern):
    class LookupSerializer(serializers.ModelSerializer):
        class Meta:
            model = LookupModel
            fields = '__all__'

    class NamespaceVersioningAPIView(generics.RetrieveUpdateDestroyAPIView):
        versioning_class = NamespaceVersioning
        serializer_class = LookupSerializer
        queryset = LookupModel.objects.all()

    # make sure regex are valid
    if path_func == re_path:
        re.compile(pattern)

    patterns_v1 = [path_func(pattern, NamespaceVersioningAPIView.as_view())]
    generator = SchemaGenerator(
        patterns=[path('v1/<int:some_param>/', include((patterns_v1, 'v1')))],
        api_version='v1',
    )
    schema = generator.get_schema(request=None, public=True)
    assert ('/v1/{some_param}/' + path_str) in schema['paths']


@pytest.mark.parametrize('viewset_cls', [AcceptHeaderVersioningViewset, AcceptHeaderVersioningViewset2])
@pytest.mark.parametrize('version', ['v1', 'v2'])
@pytest.mark.parametrize('with_request', [True, False])
def test_accept_header_versioning(no_warnings, viewset_cls, version, with_request):
    router = routers.SimpleRouter()
    router.register('x', viewset_cls, basename='x')
    generator = SchemaGenerator(
        patterns=[
            path('', include((router.urls, 'x'))),
        ],
        api_version=version,
    )
    if with_request:
        view = SpectacularAPIView(
            versioning_class=AcceptHeaderVersioning,
        )
        factory = APIRequestFactory()
        request = factory.get('x', content_type='application/vnd.oai.openapi+json')
        request = view.initialize_request(request)
    else:
        request = None
    schema = generator.get_schema(request=request, public=True)
    assert_schema(schema, f'tests/test_versioning_accept_{version}.yml')


urlpatterns_namespace = [
    path('x/', NamespaceVersioningViewset.as_view({'get': 'list'})),
    path('schema/', SpectacularAPIView.as_view(
        versioning_class=NamespaceVersioning
    ), name='schema-nv-versioned'),
    path('schema/ui', SpectacularSwaggerView.as_view(
        versioning_class=NamespaceVersioning, url_name='schema-nv-versioned'
    )),
]
urlpatterns_path = [
    path('x/', PathVersioningViewset2.as_view({'get': 'list'})),
    path('schema/', SpectacularAPIView.as_view(
        versioning_class=URLPathVersioning
    ), name='schema-pv-versioned'),
    path('schema/ui', SpectacularSwaggerView.as_view(
        versioning_class=URLPathVersioning, url_name='schema-pv-versioned'
    )),
]
urlpatterns_accept_header = [
    path('x/', AcceptHeaderVersioningViewset2.as_view({'get': 'list'})),
    path('schema/', SpectacularAPIView.as_view(
        versioning_class=AcceptHeaderVersioning
    ), name='schema-ahv-versioned'),
    path('schema/ui', SpectacularSwaggerView.as_view(
        versioning_class=AcceptHeaderVersioning, url_name='schema-ahv-versioned'
    )),
]
urlpatterns = [
    # path versioning
    re_path(r'^api/pv/(?P<version>[v1|v2]+)/', include(urlpatterns_path)),
    # namespace versioning
    path('api/nv/v1/', include((urlpatterns_namespace, 'v1'))),
    path('api/nv/v2/', include((urlpatterns_namespace, 'v2'))),
    # accept header versioning
    path('api/ahv/', include((urlpatterns_accept_header, 'x'))),
    # all unversioned
    path('api/schema/', SpectacularAPIView.as_view()),
    # manually versioned schema view that is in itself unversioned
    path('api/schema-v2/', SpectacularAPIView.as_view(api_version='v2')),
]


@pytest.mark.parametrize(['url', 'path_count'], [
    ('/api/nv/v2/schema/', 8),  # v2 nv + v2 pv + v2 ahv + unversioned
    ('/api/pv/v1/schema/', 8),  # v1 nv + v1 pv + v1 ahv + unversioned
    ('/api/schema-v2/', 8),  # v2 nv + v2 pv + v2 ahv + unversioned
    ('/api/schema/', 2),  # unversioned schema
    ('/api/schema/?version=v2', 8),  # versioned with query param
])
@pytest.mark.urls(__name__)
def test_spectacular_view_versioning(no_warnings, url, path_count):
    response = APIClient().get(url)
    assert response.status_code == 200
    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    validate_schema(schema)
    assert len(schema['paths']) == path_count


@pytest.mark.parametrize('version', ['v1', 'v2'])
@pytest.mark.urls(__name__)
def test_spectacular_view_accept_header_versioning(no_warnings, version):
    response = APIClient().get(
        '/api/ahv/schema/', HTTP_ACCEPT=f'application/json; version={version}'
    )
    assert response.status_code == 200, response.content
    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    validate_schema(schema)
    assert schema['info']['version'] == f'0.0.0 ({version})'
    assert len(schema['paths']) == 8


@pytest.mark.parametrize(['url', 'schema_url'], [
    ('/api/nv/v1/schema/ui', b'/api/nv/v1/schema/'),
    ('/api/nv/v2/schema/ui', b'/api/nv/v2/schema/'),
    ('/api/pv/v1/schema/ui', b'/api/pv/v1/schema/'),
    ('/api/pv/v2/schema/ui', b'/api/pv/v2/schema/'),
])
@pytest.mark.urls(__name__)
def test_spectacular_ui_view_versioning(no_warnings, url, schema_url):
    response = APIClient().get(url)
    assert schema_url in response.content


@pytest.mark.urls(__name__)
def test_spectacular_versioning_info_object_variations(no_warnings):
    # with default VERSION 0.0.0
    response = APIClient().get('/api/nv/v2/schema/')
    assert b'version: 0.0.0 (v2)\n' in response.content
    response = APIClient().get('/api/schema/')
    assert b'version: 0.0.0\n' in response.content

    with mock.patch('drf_spectacular.settings.spectacular_settings.VERSION', None):
        response = APIClient().get('/api/nv/v2/schema/')
        assert b'version: v2\n' in response.content
        response = APIClient().get('/api/schema/')
        assert b"version: ''\n" in response.content



from unittest import mock

import pytest
import yaml
from django import __version__ as DJANGO_VERSION
from django.http import HttpResponseRedirect
from django.urls import path
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.test import APIClient

from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import extend_schema
from drf_spectacular.validation import validate_schema
from drf_spectacular.views import (
    SpectacularAPIView, SpectacularRedocView, SpectacularSwaggerOauthRedirectView,
    SpectacularSwaggerSplitView, SpectacularSwaggerView,
)


@extend_schema(responses=OpenApiTypes.FLOAT)
@api_view(http_method_names=['GET'])
def pi(request):
    return Response(3.1415)


urlpatterns_v1 = [path('api/v1/pi/', pi)]
urlpatterns_v1.append(
    path('api/v1/schema/', SpectacularAPIView.as_view(urlconf=urlpatterns_v1))
)

urlpatterns_v2 = [
    path('api/v2/pi/', pi),
    path('api/v2/pi-fast/', pi),
    path('api/v2/schema/swagger-ui/', SpectacularSwaggerView.as_view(), name='swagger'),
    path(
        "api/v1/schema/swagger-ui/oauth2-redirect.html",
        SpectacularSwaggerOauthRedirectView.as_view(),
        name="swagger-oauth-redirect"),
    path('api/v2/schema/swagger-ui-alt/', SpectacularSwaggerSplitView.as_view(), name='swagger-alt'),
    path('api/v2/schema/redoc/', SpectacularRedocView.as_view(), name='redoc'),
]
urlpatterns_v2.append(
    path('api/v2/schema/', SpectacularAPIView.as_view(urlconf=urlpatterns_v2), name='schema'),
)
urlpatterns_str_import = [
    path('api/schema-str1/', SpectacularAPIView.as_view(urlconf=['tests.test_view']), name='schema_str1'),
    path('api/schema-str2/', SpectacularAPIView.as_view(urlconf='tests.test_view'), name='schema_str2'),
    path('api/schema-err1/', SpectacularAPIView.as_view(urlconf=['tests.error']), name='schema_err1'),
    path('api/schema-err2/', SpectacularAPIView.as_view(urlconf='tests.error'), name='schema_err2'),
]

urlpatterns = urlpatterns_v1 + urlpatterns_v2 + urlpatterns_str_import


@pytest.mark.urls(__name__)
def test_spectacular_view(no_warnings):
    response = APIClient().get('/api/v1/schema/')
    assert response.status_code == 200
    assert response.content.startswith(b'openapi: 3.0.3\n')
    assert response.accepted_media_type == 'application/vnd.oai.openapi'

    if DJANGO_VERSION > '3':
        assert response.headers.get('CONTENT-DISPOSITION') == 'inline; filename="schema.yaml"'

    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    validate_schema(schema)
    assert len(schema['paths']) == 2


@pytest.mark.urls(__name__)
def test_spectacular_view_custom_urlconf(no_warnings):
    response = APIClient().get('/api/v2/schema/')
    assert response.status_code == 200

    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    validate_schema(schema)
    assert len(schema['paths']) == 3

    response = APIClient().get('/api/v2/pi-fast/')
    assert response.status_code == 200
    assert response.content == b'3.1415'


@pytest.mark.parametrize(['accept', 'format', 'indent'], [
    ('application/vnd.oai.openapi', 'yaml', None),
    ('application/yaml', 'yaml', None),
    ('application/vnd.oai.openapi+json', 'json', 4),
    ('application/json', 'json', 4),
    ('application/json; indent=8', 'json', 8),
])
@pytest.mark.urls(__name__)
def test_spectacular_view_accept(accept, format, indent):
    response = APIClient().get('/api/v1/schema/', HTTP_ACCEPT=accept)
    assert response.status_code == 200
    assert response.accepted_media_type == accept
    if format == 'json':
        assert response.content.startswith(b'{\n' + indent * b' ' + b'"openapi": "3.0.3"')
    if format == 'yaml':
        assert response.content.startswith(b'openapi: 3.0.3\n')


@pytest.mark.urls(__name__)
def test_spectacular_view_accept_unknown(no_warnings):
    response = APIClient().get('/api/v1/schema/', HTTP_ACCEPT='application/unknown')
    assert response.status_code == 406
    assert response.content == (
        b'detail:\n  string: Could not satisfy the request Accept header.\n'
        b'  code: not_acceptable\n'
    )


@pytest.mark.parametrize('ui', ['redoc', 'swagger-ui'])
@pytest.mark.urls(__name__)
def test_spectacular_ui_view(no_warnings, ui):
    from drf_spectacular.settings import spectacular_settings
    response = APIClient().get(f'/api/v2/schema/{ui}/')
    assert response.status_code == 200
    assert response.content.startswith(b'<!DOCTYPE html>')
    if ui == 'redoc':
        assert b'<title>Redoc</title>' in response.content
        assert spectacular_settings.REDOC_DIST.encode() in response.content
    else:
        assert b'<title>Swagger</title>' in response.content
        assert spectacular_settings.SWAGGER_UI_DIST.encode() in response.content
    assert b'"/api/v2/schema/"' in response.content


@pytest.mark.urls(__name__)
def test_spectacular_swagger_ui_alternate(no_warnings):
    # first request for the html
    response = APIClient().get('/api/v2/schema/swagger-ui-alt/')
    assert response.status_code == 200
    assert response.content.startswith(b'<!DOCTYPE html>')
    assert b'"/api/v2/schema/swagger-ui-alt/?script="' in response.content
    # second request to obtain js swagger config (CSP self)
    response = APIClient().get('/api/v2/schema/swagger-ui-alt/?script=')
    assert response.status_code == 200
    assert b'"/api/v2/schema/"' in response.content


@mock.patch(
    'drf_spectacular.settings.spectacular_settings.SWAGGER_UI_SETTINGS',
    '{"deepLinking": true}'
)
@pytest.mark.urls(__name__)
def test_spectacular_ui_with_raw_settings(no_warnings):
    response = APIClient().get('/api/v2/schema/swagger-ui/')
    assert response.status_code == 200
    assert b'const swaggerSettings = {"deepLinking": true};\n' in response.content


@pytest.mark.urls(__name__)
def test_spectacular_ui_param_passthrough(no_warnings):
    response = APIClient().get('/api/v2/schema/swagger-ui/?foo=bar&lang=jp&version=v2')
    assert response.status_code == 200
    assert b'url: "/api/v2/schema/?lang\\u003Djp\\u0026version\\u003Dv2"' in response.content


@pytest.mark.parametrize('url', ['/api/schema-str1/', '/api/schema-str2/'])
@pytest.mark.urls(__name__)
def test_spectacular_urlconf_module_list_import(no_warnings, url):
    response = APIClient().get(url)
    assert response.status_code == 200
    assert b'/api/v1/pi/' in response.content
    assert b'/api/v2/pi/' in response.content


@pytest.mark.parametrize('url', ['/api/schema-err1/', '/api/schema-err2/'])
@pytest.mark.urls(__name__)
def test_spectacular_urlconf_module_list_import_error(no_warnings, url):
    with pytest.raises(ModuleNotFoundError):
        APIClient().get(url)


@pytest.mark.parametrize('get_params', ['', 'code=foobar123&state=xyz&session_state=hello-world'])
@pytest.mark.urls(__name__)
def test_swagger_oauth_redirect_view(get_params):
    # act
    response = APIClient().get('/api/v1/schema/swagger-ui/oauth2-redirect.html?' + get_params)

    # assert
    assert response.status_code == 302
    if isinstance(response, HttpResponseRedirect):
        # older django versions test client directly returns the response instance
        assert response.url == '/static/drf_spectacular_sidecar/swagger-ui-dist/oauth2-redirect.html?' + get_params
    else:
        assert response.headers['Location'] ==\
               '/static/drf_spectacular_sidecar/swagger-ui-dist/oauth2-redirect.html?' + get_params



import pytest
import yaml
from django.urls import path
from rest_framework import serializers
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.test import APIClient

from drf_spectacular.utils import extend_schema
from drf_spectacular.views import SpectacularAPIView


def custom_hook(endpoints, **kwargs):
    return [
        (path.rstrip('/'), path_regex.rstrip('/'), method, callback)
        for path, path_regex, method, callback in endpoints
    ]


class XSerializer(serializers.Serializer):
    field = serializers.CharField()


@extend_schema(request=XSerializer, responses=XSerializer)
@api_view(http_method_names=['POST'])
def pi(request):
    return Response(3.1415)  # pragma: no cover


urlpatterns = [
    path('api/pi/', pi),
    path('api/schema/', SpectacularAPIView.as_view(), name='schema'),
    path('api/schema-custom/', SpectacularAPIView.as_view(
        custom_settings={
            'TITLE': 'Custom settings with this SpectacularAPIView',
            'SCHEMA_PATH_PREFIX': '',
            'COMPONENT_SPLIT_REQUEST': True,
            'PREPROCESSING_HOOKS': ['tests.test_custom_settings.custom_hook']
        }
    ), name='schema-custom'),
    path('api/schema-invalid/', SpectacularAPIView.as_view(
        custom_settings={'INVALID': 'INVALID'}
    ), name='schema-invalid'),
    path('api/schema-invalid2/', SpectacularAPIView.as_view(
        custom_settings={'SERVE_PUBLIC': 'INVALID'}
    ), name='schema-invalid2'),
]


@pytest.mark.urls(__name__)
def test_custom_settings(no_warnings):
    response = APIClient().get('/api/schema-custom/')
    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    assert schema['info']['title']
    assert '/api/pi' in schema['paths']  # hook executed
    assert ['api'] == schema['paths']['/api/pi']['post']['tags']  # SCHEMA_PATH_PREFIX
    assert 'XRequest' in schema['components']['schemas']  # COMPONENT_SPLIT_REQUEST

    response = APIClient().get('/api/schema/')
    schema = yaml.load(response.content, Loader=yaml.SafeLoader)
    assert not schema['info']['title']
    assert '/api/pi/' in schema['paths']  # hook not executed
    assert ['pi'] == schema['paths']['/api/pi/']['post']['tags']  # SCHEMA_PATH_PREFIX
    assert 'XRequest' not in schema['components']['schemas']  # COMPONENT_SPLIT_REQUEST


@pytest.mark.urls(__name__)
def test_invalid_custom_settings():
    with pytest.raises(AttributeError):
        APIClient().get('/api/schema-invalid/')
    with pytest.raises(AttributeError):
        APIClient().get('/api/schema-invalid2/')



from drf_rw_serializers.generics import GenericAPIView as RWGenericAPIView

from drf_spectacular.openapi import AutoSchema


class CustomAutoSchema(AutoSchema):
    """ Utilize custom drf_rw_serializers methods for directional serializers """

    def get_request_serializer(self):
        if isinstance(self.view, RWGenericAPIView):
            return self.view.get_write_serializer()
        return self._get_serializer()

    def get_response_serializers(self):
        if isinstance(self.view, RWGenericAPIView):
            return self.view.get_read_serializer()
        return self._get_serializer()



from drf_spectacular.extensions import OpenApiAuthenticationExtension
from drf_spectacular.plumbing import build_bearer_security_scheme_object

class AdfsAccessTokenAuthenticationScheme(OpenApiAuthenticationExtension):
    target_class = 'django_auth_adfs.rest_framework.AdfsAccessTokenAuthentication'
    name = 'jwtAuth'

    def get_security_definition(self, auto_schema):
        return build_bearer_security_scheme_object(header_name='AUTHORIZATION',
                                                   token_prefix='Bearer',
                                                   bearer_format='JWT')



import sys
from typing import List
from unittest import mock

import pytest
from rest_framework import mixins, serializers, viewsets
from rest_framework.decorators import action

# Use 3.9 instead of 3.8 version because of missing __required_keys__ feature,
# which allows for a consistent test schema over different versions.
if sys.version_info >= (3, 9):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

from drf_spectacular.contrib.djangorestframework_camel_case import camelize_serializer_fields
from drf_spectacular.utils import extend_schema
from tests import assert_schema, generate_schema


class NestedObject2(TypedDict):
    field_seven: int
    field_eight: str


class NestedObject(TypedDict):
    field_three: int
    field_four: str
    field_five: NestedObject2
    field_six: List[NestedObject2]


class FakeSerializer(serializers.Serializer):
    field_one = serializers.CharField()
    field_two = serializers.CharField()

    field_nested = serializers.SerializerMethodField()

    def get_field_nested(self) -> NestedObject:  # type: ignore
        pass  # pragma: no cover


class FakeViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = FakeSerializer

    @extend_schema(responses=FakeSerializer)
    @action(detail=False, serializer_class=FakeSerializer)
    def home(self, request):
        ...  # pragma: no cover


@mock.patch(
    'drf_spectacular.settings.spectacular_settings.POSTPROCESSING_HOOKS',
    [camelize_serializer_fields]
)
@pytest.mark.contrib('djangorestframework_camel_case')
def test_camelize_serializer_fields():
    assert_schema(
        generate_schema('a_b_c', FakeViewset),
        'tests/contrib/test_djangorestframework_camel_case.yml'
    )



import inspect
import sys
from typing import Any, Callable, Dict, Optional, Sequence, Tuple, Type, TypeVar, Union

# direct import due to https://github.com/microsoft/pyright/issues/3025
if sys.version_info >= (3, 8):
    from typing import Final, Literal
else:
    from typing_extensions import Final, Literal

from rest_framework.fields import Field, empty
from rest_framework.serializers import Serializer
from rest_framework.settings import api_settings

from drf_spectacular.drainage import (
    error, get_view_method_names, isolate_view_method, set_override, warn,
)
from drf_spectacular.types import OpenApiTypes, _KnownPythonTypes

_SerializerType = Union[Serializer, Type[Serializer]]
_FieldType = Union[Field, Type[Field]]
_ParameterLocationType = Literal['query', 'path', 'header', 'cookie']
Direction = Literal['request', 'response']


class PolymorphicProxySerializer(Serializer):
    """
    This class is to be used with :func:`@extend_schema <.extend_schema>` to
    signal a request/response might be polymorphic (accepts/returns data
    possibly from different serializers). Usage usually looks like this:

    .. code-block::

        @extend_schema(
            request=PolymorphicProxySerializer(
                component_name='MetaPerson',
                serializers=[
                    LegalPersonSerializer, NaturalPersonSerializer,
                ],
                resource_type_field_name='person_type',
            )
        )
        def create(self, request, *args, **kwargs):
            return Response(...)

    **Beware** that this is not a real serializer and it will raise an AssertionError
    if used in that way. It **cannot** be used in views as ``serializer_class``
    or as field in an actual serializer. It is solely meant for annotation purposes.

    Also make sure that each sub-serializer has a field named after the value of
    ``resource_type_field_name`` (discriminator field). Generated clients will likely
    depend on the existence of this field.

    Setting ``resource_type_field_name`` to ``None`` will remove the discriminator
    altogether. This may be useful in certain situations, but will most likely break
    client generation. Another use-case is explicit control over sub-serializer's ``many``
    attribute. To explicitly control this aspect, you need disable the discriminator with
    ``resource_type_field_name=None`` as well as disable automatic list handling with
    ``many=False``.

    It is **strongly** recommended to pass the ``Serializers`` as **list**,
    and by that let *drf-spectacular* retrieve the field and handle the mapping
    automatically. In special circumstances, the field may not available when
    *drf-spectacular* processes the serializer. In those cases you can explicitly state
    the mapping with ``{'legal': LegalPersonSerializer, ...}``, but it is then your
    responsibility to have a valid mapping.
    """
    def __init__(
            self,
            component_name: str,
            serializers: Union[Sequence[_SerializerType], Dict[str, _SerializerType]],
            resource_type_field_name: Optional[str],
            many: Optional[bool] = None,
    ):
        self.component_name = component_name
        self.serializers = serializers
        self.resource_type_field_name = resource_type_field_name
        if self._many is False:  # type: ignore[attr-defined]
            set_override(self, 'many', False)
        super().__init__()

    def __new__(cls, *args, **kwargs):
        many = kwargs.pop('many', None)
        if many is True:
            instance = cls.many_init(*args, **kwargs)
        else:
            instance = super().__new__(cls, *args, **kwargs)
        instance._many = many
        return instance

    @property
    def data(self):
        self._trap()

    def to_internal_value(self, data):
        self._trap()

    def to_representation(self, instance):
        self._trap()

    def _trap(self):
        raise AssertionError(
            "PolymorphicProxySerializer is an annotation helper and not supposed to "
            "be used for real requests. See documentation for correct usage."
        )


class OpenApiSchemaBase:
    pass


class OpenApiExample(OpenApiSchemaBase):
    """
    Helper class to document a API parameter / request body / response body
    with a concrete example value.

    It is recommended to provide a singular example value, since pagination
    and list responses are handled by drf-spectacular.

    The example will be attached to the operation object where appropriate,
    i.e. where the given ``media_type``, ``status_code`` and modifiers match.
    Example that do not match any scenario are ignored.
    """
    def __init__(
            self,
            name: str,
            value: Any = None,
            external_value: str = '',
            summary: str = '',
            description: str = '',
            request_only: bool = False,
            response_only: bool = False,
            parameter_only: Optional[Tuple[str, _ParameterLocationType]] = None,
            media_type: str = 'application/json',
            status_codes: Optional[Sequence[Union[str, int]]] = None,
    ):
        self.name = name
        self.summary = summary
        self.description = description
        self.value = value
        self.external_value = external_value
        self.request_only = request_only
        self.response_only = response_only
        self.parameter_only = parameter_only
        self.media_type = media_type
        self.status_codes = list(map(str, status_codes)) if status_codes else ['200', '201']


class OpenApiParameter(OpenApiSchemaBase):
    """
    Helper class to document request query/path/header/cookie parameters.
    Can also be used to document response headers.

    Please note that not all arguments apply to all ``location``/``type``/direction
    variations, e.g. path parameters are ``required=True`` by definition.

    For valid ``style`` choices please consult the
    `OpenAPI specification <https://swagger.io/specification/#style-values>`_.
    """
    QUERY: Final = 'query'
    PATH: Final = 'path'
    HEADER: Final = 'header'
    COOKIE: Final = 'cookie'

    def __init__(
            self,
            name: str,
            type: Union[_SerializerType, _KnownPythonTypes, OpenApiTypes, dict] = str,
            location: _ParameterLocationType = QUERY,
            required: bool = False,
            description: str = '',
            enum: Optional[Sequence[Any]] = None,
            pattern: Optional[str] = None,
            deprecated: bool = False,
            style: Optional[str] = None,
            explode: Optional[bool] = None,
            default: Any = None,
            allow_blank: bool = True,
            many: Optional[bool] = None,
            examples: Optional[Sequence[OpenApiExample]] = None,
            extensions: Optional[Dict[str, Any]] = None,
            exclude: bool = False,
            response: Union[bool, Sequence[Union[int, str]]] = False,
    ):
        self.name = name
        self.type = type
        self.location = location
        self.required = required
        self.description = description
        self.enum = enum
        self.pattern = pattern
        self.deprecated = deprecated
        self.style = style
        self.explode = explode
        self.default = default
        self.allow_blank = allow_blank
        self.many = many
        self.examples = examples or []
        self.extensions = extensions
        self.exclude = exclude
        self.response = response


class OpenApiResponse(OpenApiSchemaBase):
    """
    Helper class to bundle a response object (``Serializer``, ``OpenApiType``,
    raw schema, etc) together with a response object description and/or examples.
    Examples can alternatively be provided via :func:`@extend_schema <.extend_schema>`.

    This class is especially helpful for explicitly describing status codes on a
    "Response Object" level.
    """
    def __init__(
            self,
            response: Any = None,
            description: str = '',
            examples: Optional[Sequence[OpenApiExample]] = None
    ):
        self.response = response
        self.description = description
        self.examples = examples or []


F = TypeVar('F', bound=Callable[..., Any])


class OpenApiCallback(OpenApiSchemaBase):
    """
    Helper class to bundle a callback definition. This specifies a view on the callee's
    side, effectively stating the expectations on the receiving end. Please note that this
    particular :func:`@extend_schema <.extend_schema>` instance operates from the perspective
    of the callback origin, which means that ``request`` specifies the outgoing request.

    For convenience sake, we assume the callback sends ``application/json`` and return a ``200``.
    If that is not sufficient, you can use ``request`` and ``responses`` overloads just as you
    normally would.

    :param name: Name under which the this callback is listed in the schema.
    :param path: Path on which the callback operation is performed. To reference request
        body contents, please refer to OpenAPI specification's
        `key expressions <https://swagger.io/specification/#key-expression>`_ for valid choices.
    :param decorator: :func:`@extend_schema <.extend_schema>` decorator that specifies the receiving
        endpoint. In this special context the allowed parameters are ``requests``, ``responses``,
        ``summary``, ``description``, ``deprecated``.
    """
    def __init__(
            self,
            name: str,
            path: str,
            decorator: Union[Callable[[F], F], Dict[str, Callable[[F], F]], Dict[str, Any]],
    ):
        self.name = name
        self.path = path
        self.decorator = decorator


def extend_schema(
        operation_id: Optional[str] = None,
        parameters: Optional[Sequence[Union[OpenApiParameter, _SerializerType]]] = None,
        request: Any = empty,
        responses: Any = empty,
        auth: Optional[Sequence[str]] = None,
        description: Optional[str] = None,
        summary: Optional[str] = None,
        deprecated: Optional[bool] = None,
        tags: Optional[Sequence[str]] = None,
        filters: Optional[bool] = None,
        exclude: bool = False,
        operation: Optional[Dict] = None,
        methods: Optional[Sequence[str]] = None,
        versions: Optional[Sequence[str]] = None,
        examples: Optional[Sequence[OpenApiExample]] = None,
        extensions: Optional[Dict[str, Any]] = None,
        callbacks: Optional[Sequence[OpenApiCallback]] = None,
        external_docs: Optional[Union[Dict[str, str], str]] = None,
) -> Callable[[F], F]:
    """
    Decorator mainly for the "view" method kind. Partially or completely overrides
    what would be otherwise generated by drf-spectacular.

    :param operation_id: replaces the auto-generated operation_id. make sure there
        are no naming collisions.
    :param parameters: list of additional or replacement parameters added to the
        auto-discovered fields.
    :param responses: replaces the discovered Serializer. Takes a variety of
        inputs that can be used individually or combined

        - ``Serializer`` class
        - ``Serializer`` instance (e.g. ``Serializer(many=True)`` for listings)
        - basic types or instances of ``OpenApiTypes``
        - :class:`.OpenApiResponse` for bundling any of the other choices together with
          either a dedicated response description and/or examples.
        - :class:`.PolymorphicProxySerializer` for signaling that
          the operation may yield data from different serializers depending
          on the circumstances.
        - ``dict`` with status codes as keys and one of the above as values.
          Additionally in this case, it is also possible to provide a raw schema dict
          as value.
        - ``dict`` with tuples (status_code, media_type) as keys and one of the above
          as values. Additionally in this case, it is also possible to provide a raw
          schema dict as value.
    :param request: replaces the discovered ``Serializer``. Takes a variety of inputs

        - ``Serializer`` class/instance
        - basic types or instances of ``OpenApiTypes``
        - :class:`.PolymorphicProxySerializer` for signaling that the operation
          accepts a set of different types of objects.
        - ``dict`` with media_type as keys and one of the above as values. Additionally in
          this case, it is also possible to provide a raw schema dict as value.
    :param auth: replace discovered auth with explicit list of auth methods
    :param description: replaces discovered doc strings
    :param summary: an optional short summary of the description
    :param deprecated: mark operation as deprecated
    :param tags: override default list of tags
    :param filters: ignore list detection and forcefully enable/disable filter discovery
    :param exclude: set True to exclude operation from schema
    :param operation: manually override what auto-discovery would generate. you must
        provide a OpenAPI3-compliant dictionary that gets directly translated to YAML.
    :param methods: scope extend_schema to specific methods. matches all by default.
    :param versions: scope extend_schema to specific API version. matches all by default.
    :param examples: attach request/response examples to the operation
    :param extensions: specification extensions, e.g. ``x-badges``, ``x-code-samples``, etc.
    :param callbacks: associate callbacks with this endpoint
    :param external_docs: Link external documentation. Provide a dict with an "url" key and
        optionally a "description" key. For convenience, if only a string is given it is
        treated as the URL.
    :return:
    """
    if methods is not None:
        methods = [method.upper() for method in methods]

    def decorator(f):
        BaseSchema = (
            # explicit manually set schema or previous view annotation
            getattr(f, 'schema', None)
            # previously set schema with @extend_schema on views methods
            or getattr(f, 'kwargs', {}).get('schema', None)
            # previously set schema with @extend_schema on @api_view
            or getattr(getattr(f, 'cls', None), 'kwargs', {}).get('schema', None)
            # the default
            or api_settings.DEFAULT_SCHEMA_CLASS
        )

        if not inspect.isclass(BaseSchema):
            BaseSchema = BaseSchema.__class__

        def is_in_scope(ext_schema):
            version, _ = ext_schema.view.determine_version(
                ext_schema.view.request,
                **ext_schema.view.kwargs
            )
            version_scope = versions is None or version in versions
            method_scope = methods is None or ext_schema.method in methods
            return method_scope and version_scope

        class ExtendedSchema(BaseSchema):
            def get_operation(self, path, path_regex, path_prefix, method, registry):
                self.method = method.upper()

                if exclude and is_in_scope(self):
                    return None
                if operation is not None and is_in_scope(self):
                    return operation
                return super().get_operation(path, path_regex, path_prefix, method, registry)

            def get_operation_id(self):
                if operation_id and is_in_scope(self):
                    return operation_id
                return super().get_operation_id()

            def get_override_parameters(self):
                if parameters and is_in_scope(self):
                    return super().get_override_parameters() + parameters
                return super().get_override_parameters()

            def get_auth(self):
                if auth is not None and is_in_scope(self):
                    return auth
                return super().get_auth()

            def get_examples(self):
                if examples and is_in_scope(self):
                    return super().get_examples() + examples
                return super().get_examples()

            def get_request_serializer(self):
                if request is not empty and is_in_scope(self):
                    return request
                return super().get_request_serializer()

            def get_response_serializers(self):
                if responses is not empty and is_in_scope(self):
                    return responses
                return super().get_response_serializers()

            def get_description(self):
                if description and is_in_scope(self):
                    return description
                return super().get_description()

            def get_summary(self):
                if summary and is_in_scope(self):
                    return str(summary)
                return super().get_summary()

            def is_deprecated(self):
                if deprecated and is_in_scope(self):
                    return deprecated
                return super().is_deprecated()

            def get_tags(self):
                if tags is not None and is_in_scope(self):
                    return tags
                return super().get_tags()

            def get_extensions(self):
                if extensions and is_in_scope(self):
                    return extensions
                return super().get_extensions()

            def get_filter_backends(self):
                if filters is not None and is_in_scope(self):
                    return getattr(self.view, 'filter_backends', []) if filters else []
                return super().get_filter_backends()

            def get_callbacks(self):
                if callbacks is not None and is_in_scope(self):
                    return callbacks
                return super().get_callbacks()

            def get_external_docs(self):
                if external_docs is not None and is_in_scope(self):
                    return external_docs
                return super().get_external_docs()

        if inspect.isclass(f):
            # either direct decoration of views, or unpacked @api_view from OpenApiViewExtension
            if operation_id is not None or operation is not None:
                error(
                    f'using @extend_schema on viewset class {f.__name__} with parameters '
                    f'operation_id or operation will most likely result in a broken schema.',
                    delayed=f,
                )
            # reorder schema class MRO so that view method annotation takes precedence
            # over view class annotation. only relevant if there is a method annotation
            for view_method_name in get_view_method_names(view=f, schema=BaseSchema):
                if 'schema' not in getattr(getattr(f, view_method_name), 'kwargs', {}):
                    continue
                view_method = isolate_view_method(f, view_method_name)
                view_method.kwargs['schema'] = type(
                    'ExtendedMetaSchema', (view_method.kwargs['schema'], ExtendedSchema), {}
                )
            # persist schema on class to provide annotation to derived view methods.
            # the second purpose is to serve as base for view multi-annotation
            f.schema = ExtendedSchema()
            return f
        elif callable(f) and hasattr(f, 'cls'):
            # 'cls' attr signals that as_view() was called, which only applies to @api_view.
            # keep a "unused" schema reference at root level for multi annotation convenience.
            setattr(f.cls, 'kwargs', {'schema': ExtendedSchema})
            # set schema on method kwargs context to emulate regular view behaviour.
            for method in f.cls.http_method_names:
                setattr(getattr(f.cls, method), 'kwargs', {'schema': ExtendedSchema})
            return f
        elif callable(f):
            # custom actions have kwargs in their context, others don't. create it so our create_view
            # implementation can overwrite the default schema
            if not hasattr(f, 'kwargs'):
                f.kwargs = {}
            # this simulates what @action is actually doing. somewhere along the line in this process
            # the schema is picked up from kwargs and used. it's involved my dear friends.
            # use class instead of instance due to descriptor weakref reverse collisions
            f.kwargs['schema'] = ExtendedSchema
            return f
        else:
            return f

    return decorator


def extend_schema_field(
        field: Union[_SerializerType, _FieldType, OpenApiTypes, Dict],
        component_name: Optional[str] = None
) -> Callable[[F], F]:
    """
    Decorator for the "field" kind. Can be used with ``SerializerMethodField`` (annotate the actual
    method) or with custom ``serializers.Field`` implementations.

    If your custom serializer field base class is already the desired type, decoration is not necessary.
    To override the discovered base class type, you can decorate your custom field class.

    Always takes precedence over other mechanisms (e.g. type hints, auto-discovery).

    :param field: accepts a ``Serializer``, :class:`~.types.OpenApiTypes` or raw ``dict``
    :param component_name: signals that the field should be broken out as separate component
    """

    def decorator(f):
        set_override(f, 'field', field)
        set_override(f, 'field_component_name', component_name)
        return f

    return decorator


def extend_schema_serializer(
        many: Optional[bool] = None,
        exclude_fields: Optional[Sequence[str]] = None,
        deprecate_fields: Optional[Sequence[str]] = None,
        examples: Optional[Sequence[OpenApiExample]] = None,
        extensions: Optional[Dict[str, Any]] = None,
        component_name: Optional[str] = None,
) -> Callable[[F], F]:
    """
    Decorator for the "serializer" kind. Intended for overriding default serializer behaviour that
    cannot be influenced through :func:`@extend_schema <.extend_schema>`.

    :param many: override how serializer is initialized. Mainly used to coerce the list view detection
        heuristic to acknowledge a non-list serializer.
    :param exclude_fields: fields to ignore while processing the serializer. only affects the
        schema. fields will still be exposed through the API.
    :param deprecate_fields: fields to mark as deprecated while processing the serializer.
    :param examples: define example data to serializer.
    :param extensions: specification extensions, e.g. ``x-is-dynamic``, etc.
    :param component_name: override default class name extraction.
    """
    def decorator(klass):
        if many is not None:
            set_override(klass, 'many', many)
        if exclude_fields:
            set_override(klass, 'exclude_fields', exclude_fields)
        if deprecate_fields:
            set_override(klass, 'deprecate_fields', deprecate_fields)
        if examples:
            set_override(klass, 'examples', examples)
        if extensions:
            set_override(klass, 'extensions', extensions)
        if component_name:
            set_override(klass, 'component_name', component_name)
        return klass

    return decorator


def extend_schema_view(**kwargs) -> Callable[[F], F]:
    """
    Convenience decorator for the "view" kind. Intended for annotating derived view methods that
    are are not directly present in the view (usually methods like ``list`` or ``retrieve``).
    Spares you from overriding methods like ``list``, only to perform a super call in the body
    so that you have have something to attach :func:`@extend_schema <.extend_schema>` to.

    This decorator also takes care of safely attaching annotations to derived view methods,
    preventing leakage into unrelated views.

    :param kwargs: method names as argument names and :func:`@extend_schema <.extend_schema>`
      calls as values
    """
    def decorator(view):
        # special case for @api_view. redirect decoration to enclosed WrappedAPIView
        if callable(view) and hasattr(view, 'cls'):
            extend_schema_view(**kwargs)(view.cls)
            return view

        available_view_methods = get_view_method_names(view)

        for method_name, method_decorator in kwargs.items():
            if method_name not in available_view_methods:
                warn(
                    f'@extend_schema_view argument "{method_name}" was not found on view '
                    f'{view.__name__}. method override for "{method_name}" will be ignored.',
                    delayed=view
                )
                continue

            # the context of derived methods must not be altered, as it belongs to the
            # other view. create a new context so the schema can be safely stored in the
            # wrapped_method. view methods that are not derived can be safely altered.
            if hasattr(method_decorator, '__iter__'):
                for sub_method_decorator in method_decorator:
                    sub_method_decorator(isolate_view_method(view, method_name))
            else:
                method_decorator(isolate_view_method(view, method_name))
        return view

    return decorator


def inline_serializer(name: str, fields: Dict[str, Field], **kwargs) -> Serializer:
    """
    A helper function to create an inline serializer. Primary use is with
    :func:`@extend_schema <.extend_schema>`, where one needs an implicit one-off
    serializer that is not reflected in an actual class.

    :param name: name of the
    :param fields: dict with field names as keys and serializer fields as values
    :param kwargs: optional kwargs for serializer initialization
    """
    serializer_class = type(name, (Serializer,), fields)
    return serializer_class(**kwargs)



import datetime
import json
import os
import re
import subprocess

envs = subprocess.check_output(['tox', '-l']).decode().rstrip().split('\n')
matrix = []

for env in envs:
    version = re.search(r'^py(?P<major>\d)(?P<minor>\d+)-', env)

    # github "commit" checks will fail even though workflow passes overall.
    # temp remove the optional targets to make github CI work.
    if 'master' in env:
        continue

    matrix.append({
        'toxenv': env,
        'python-version': f'{version.group("major")}.{version.group("minor")}',
        'experimental': bool('master' in env)
    })

with open(os.environ['GITHUB_OUTPUT'], "a") as fh:
    fh.write(f"date={datetime.date.today()}\n")
    fh.write(f"matrix={json.dumps(matrix)}\n")



from unittest import mock

import pytest
from django import __version__ as DJANGO_VERSION
from django.urls import path
from rest_framework import mixins, routers, serializers, viewsets

from tests import assert_schema, generate_schema

try:
    from rest_framework_jwt.authentication import JSONWebTokenAuthentication
except ImportError:
    JSONWebTokenAuthentication = None


class XSerializer(serializers.Serializer):
    uuid = serializers.UUIDField()


class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
    serializer_class = XSerializer
    authentication_classes = [JSONWebTokenAuthentication]
    required_scopes = ['x:read', 'x:write']


@pytest.mark.skipif(DJANGO_VERSION >= '4', reason='1.19.1 broken for Django 4.0')
@pytest.mark.contrib('rest_framework_jwt')
def test_drf_jwt(no_warnings):
    from rest_framework_jwt.views import obtain_jwt_token

    router = routers.SimpleRouter()
    router.register('x', XViewset, basename="x")

    urlpatterns = [
        *router.urls,
        path('api-token-auth/', obtain_jwt_token, name='get_token'),
    ]

    schema = generate_schema(None, patterns=urlpatterns)

    assert_schema(schema, 'tests/contrib/test_drf_jwt.yml')


@pytest.mark.skipif(DJANGO_VERSION >= '4', reason='1.19.1 broken for Django 4.0')
@pytest.mark.contrib('rest_framework_jwt')
@mock.patch('rest_framework_jwt.settings.api_settings.JWT_AUTH_HEADER_PREFIX', 'JWT')
def test_drf_jwt_non_bearer_keyword(no_warnings):
    schema = generate_schema('/x', XViewset)
    assert schema['components']['securitySchemes'] == {
        'jwtAuth': {
            'type': 'apiKey',
            'in': 'header',
            'name': 'Authorization',
            'description': 'Token-based authentication with required prefix "JWT"'
        },
    }



from unittest import mock

import pytest
from django.db import models
from rest_framework import __version__ as DRF_VERSION  # type: ignore[attr-defined]
from rest_framework import mixins, routers, serializers, viewsets

from drf_spectacular.utils import extend_schema_serializer
from tests import assert_schema, generate_schema


@pytest.mark.contrib('rest_framework_gis')
@pytest.mark.skipif(DRF_VERSION < '3.12', reason='DRF pagination schema broken')
@mock.patch('drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES', {})
def test_rest_framework_gis(no_warnings, clear_caches):
    from django.contrib.gis.db.models import (
        GeometryCollectionField, GeometryField, LineStringField, MultiLineStringField,
        MultiPointField, MultiPolygonField, PointField, PolygonField,
    )
    from rest_framework_gis.fields import GeometryField as SerializerGeometryField
    from rest_framework_gis.pagination import GeoJsonPagination
    from rest_framework_gis.serializers import GeoFeatureModelSerializer

    class GeoModel(models.Model):
        field_random1 = models.CharField(max_length=32)
        field_random2 = models.IntegerField()
        field_gis_plain = PointField()

        field_polygon = PolygonField()
        field_point = PointField()
        field_linestring = LineStringField()
        field_geometry = GeometryField()
        field_multipolygon = MultiPolygonField()
        field_multipoint = MultiPointField()
        field_multilinestring = MultiLineStringField()
        field_geometrycollection = GeometryCollectionField()

    class GeoModel2(models.Model):
        related_model = models.OneToOneField(GeoModel, on_delete=models.DO_NOTHING)
        field_point = PointField()

    router = routers.SimpleRouter()

    # all GIS fields as GeoJSON in singular and list form
    fields = [
        'Point', 'Polygon', 'Linestring', 'Geometry',
        'Multipoint', 'Multipolygon', 'Multilinestring', 'Geometrycollection'
    ]
    for name in fields:
        @extend_schema_serializer(component_name=name)
        class XSerializer(GeoFeatureModelSerializer):
            class Meta:
                model = GeoModel
                geo_field = f'field_{name.lower()}'
                auto_bbox = name == 'Polygon'
                fields = ['id', 'field_random1', 'field_random2', 'field_gis_plain']

        class XViewset(mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet):
            serializer_class = XSerializer
            queryset = GeoModel.objects.none()

        router.register(name.lower(), XViewset, basename=name)

    # plain serializer with GIS fields but without restructured container object
    class PlainSerializer(serializers.ModelSerializer):
        field_gis_related = SerializerGeometryField(source="geomodel2.field_point")

        class Meta:
            model = GeoModel
            fields = ['id', 'field_random1', 'field_random2', 'field_gis_plain', 'field_gis_related']

    class PlainViewset(mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = PlainSerializer
        queryset = GeoModel.objects.none()

    router.register('plain', PlainViewset, basename='plain')

    # GIS specific pagination
    class PlainViewset(mixins.RetrieveModelMixin, mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = PlainSerializer
        queryset = GeoModel.objects.none()
        pagination_class = GeoJsonPagination

    router.register('paginated', PlainViewset, basename='paginated')

    assert_schema(
        generate_schema(None, patterns=router.urls),
        'tests/contrib/test_rest_framework_gis.yml'
    )


@pytest.mark.contrib('rest_framework_gis', 'django_filter')
@mock.patch('drf_spectacular.settings.spectacular_settings.ENUM_NAME_OVERRIDES', {})
def test_geo_filter_set(no_warnings):
    from django.contrib.gis.db.models import PointField
    from django_filters import filters
    from django_filters.rest_framework import DjangoFilterBackend
    from rest_framework_gis.filters import GeometryFilter
    from rest_framework_gis.filterset import GeoFilterSet

    class GeoRegionModel(models.Model):
        slug = models.CharField(max_length=32)
        geom = PointField()

    class RegionFilter(GeoFilterSet):
        slug = filters.CharFilter(field_name='slug', lookup_expr='istartswith')
        contains_geom = GeometryFilter(field_name='geom', lookup_expr='contains')

        class Meta:
            model = GeoRegionModel
            fields = ['slug', 'contains_geom']

    class XSerializer(serializers.Serializer):
        slug = serializers.CharField()

    class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = XSerializer
        queryset = GeoRegionModel.objects.none()
        filterset_class = RegionFilter
        filter_backends = [DjangoFilterBackend]

    schema = generate_schema('/x', XViewset)
    assert schema['paths']['/x/']['get']['parameters'] == [
        {'in': 'query', 'name': 'contains_geom', 'schema': {'type': 'string'}},
        {'in': 'query', 'name': 'slug', 'schema': {'type': 'string'}}
    ]



from typing import TYPE_CHECKING
from unittest import mock

from rest_framework import fields, mixins, permissions, serializers, viewsets
from rest_framework.authentication import BaseAuthentication
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.views import APIView

from drf_spectacular.extensions import (
    OpenApiAuthenticationExtension, OpenApiSerializerExtension, OpenApiSerializerFieldExtension,
    OpenApiViewExtension,
)
from drf_spectacular.plumbing import (
    ResolvedComponent, build_array_type, build_basic_type, build_object_type,
)
from drf_spectacular.types import OpenApiTypes
from drf_spectacular.utils import Direction, extend_schema
from tests import generate_schema, get_response_schema
from tests.models import SimpleModel, SimpleSerializer

if TYPE_CHECKING:
    from drf_spectacular.openapi import AutoSchema


class Base64Field(fields.Field):
    pass  # pragma: no cover


def test_serializer_field_extension(no_warnings):
    class Base64FieldExtension(OpenApiSerializerFieldExtension):
        target_class = 'tests.test_extensions.Base64Field'

        def map_serializer_field(self, auto_schema, direction):
            return build_basic_type(OpenApiTypes.BYTE)

    class XSerializer(serializers.Serializer):
        hash = Base64Field()

    class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = XSerializer

    schema = generate_schema('x', XViewset)
    assert schema['components']['schemas']['X']['properties']['hash']['type'] == 'string'
    assert schema['components']['schemas']['X']['properties']['hash']['format'] == 'byte'


def test_serializer_field_extension_can_return_none(no_warnings):
    """Field extensions can return None, which should exclude them from schema"""

    class BlindBase64Field(fields.Field):
        pass  # pragma: no cover

    class Base64FieldExtension(OpenApiSerializerFieldExtension):
        target_class = BlindBase64Field

        def map_serializer_field(self, auto_schema, direction):
            return None

    # At least 1 field is required to generate schema, include 'other'
    class XSerializer(serializers.Serializer):
        hash = BlindBase64Field()
        other = fields.CharField()

    class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = XSerializer

    schema = generate_schema('x', XViewset)

    # field 'hash' is missing from the schema output
    assert schema['components']['schemas']['X'] == {
        'type': 'object', 'properties': {'other': {'type': 'string'}}, 'required': ['other']
    }


class Base32Field(fields.Field):
    pass  # pragma: no cover


def test_serializer_field_extension_with_breakout(no_warnings):
    class Base32FieldExtension(OpenApiSerializerFieldExtension):
        target_class = 'tests.test_extensions.Base32Field'

        def get_name(self):
            return 'FieldComponentName'

        def map_serializer_field(self, auto_schema, direction):
            return build_basic_type(OpenApiTypes.BYTE)

    class XSerializer(serializers.Serializer):
        hash = Base32Field()

    class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = XSerializer

    schema = generate_schema('x', XViewset)
    assert schema['components']['schemas']['FieldComponentName'] == {
        'type': 'string', 'format': 'byte'
    }


class XView(APIView):
    """ underspecified library view """
    def get(self):
        """ docstring for GET """
        return Response(1.234)  # pragma: no cover


def test_view_extension(no_warnings):
    class FixXView(OpenApiViewExtension):
        target_class = 'tests.test_extensions.XView'

        def view_replacement(self):
            class Fixed(self.target_class):
                @extend_schema(responses=OpenApiTypes.FLOAT)
                def get(self, request):
                    pass  # pragma: no cover

            return Fixed

    schema = generate_schema('x', view=XView)
    operation = schema['paths']['/x']['get']
    assert get_response_schema(operation)['type'] == 'number'
    assert operation['description'] == 'docstring for GET'


@api_view()
def x_view_function():
    """ underspecified library view """
    return Response(1.234)  # pragma: no cover


def test_view_function_extension(no_warnings):
    class FixXFunctionView(OpenApiViewExtension):
        target_class = 'tests.test_extensions.x_view_function'

        def view_replacement(self):
            fixed = extend_schema(responses=OpenApiTypes.FLOAT)(self.target_class)
            return fixed

    schema = generate_schema('x', view_function=x_view_function)
    operation = schema['paths']['/x']['get']
    assert get_response_schema(operation)['type'] == 'number'
    assert operation['description'].strip() == 'underspecified library view'


def test_extension_not_found_for_installed_app(capsys):
    class FixXFunctionView(OpenApiViewExtension):
        target_class = 'tests.test_extensions.NotExistingClass'

        def view_replacement(self):
            pass  # pragma: no cover

    OpenApiViewExtension.get_match(object())
    assert 'target class was not found' in capsys.readouterr().err


class MultiHeaderAuth(BaseAuthentication):
    pass


def test_multi_auth_scheme_extension(no_warnings):
    class MultiHeaderAuthExtension(OpenApiAuthenticationExtension):
        target_class = 'tests.test_extensions.MultiHeaderAuth'
        name = ['apiKey', 'appId']

        def get_security_definition(self, auto_schema):
            return [
                {'type': 'apiKey', 'in': 'header', 'name': 'X-API-KEY'},
                {'type': 'apiKey', 'in': 'header', 'name': 'X-APP-ID'},
            ]

    class XViewset(viewsets.ReadOnlyModelViewSet):
        queryset = SimpleModel.objects.none()
        serializer_class = SimpleSerializer
        authentication_classes = [MultiHeaderAuth]
        permission_classes = [permissions.IsAuthenticated]

    schema = generate_schema('/x', XViewset)
    assert schema['components']['securitySchemes'] == {
        'apiKey': {'type': 'apiKey', 'in': 'header', 'name': 'X-API-KEY'},
        'appId': {'type': 'apiKey', 'in': 'header', 'name': 'X-APP-ID'}
    }
    assert schema['paths']['/x/']['get']['security'] == [{'apiKey': [], 'appId': []}]


def test_serializer_list_extension(no_warnings):

    class CustomListSerializer(serializers.ListSerializer):
        def to_representation(self, data):
            return {'foo': 1, 'data': super().to_representation(data)}   # pragma: no cover

    # ListSerializer can be injected either via Meta attribute or by overriding many_init()
    class XSerializer(serializers.ModelSerializer):
        class Meta:
            model = SimpleModel
            fields = '__all__'
            list_serializer_class = CustomListSerializer

    class CustomListExtension(OpenApiSerializerExtension):
        target_class = CustomListSerializer

        def map_serializer(self, auto_schema, direction):
            component = auto_schema.resolve_serializer(self.target.child, direction)
            schema = build_object_type(
                properties={'foo': build_basic_type(int), 'data': build_array_type(component.ref)}
            )
            list_component = ResolvedComponent(
                name=f'{component.name}List',
                type=ResolvedComponent.SCHEMA,
                object=self.target.child,
                schema=schema
            )
            auto_schema.registry.register_on_missing(list_component)
            return list_component.ref

    class XViewset(mixins.ListModelMixin, viewsets.GenericViewSet):
        serializer_class = XSerializer

    schema = generate_schema('x', XViewset)
    op_schema = get_response_schema(schema['paths']['/x/']['get'])
    assert op_schema == {'$ref': '#/components/schemas/XList'}
    assert schema['components']['schemas']['XList'] == {
        'type': 'object',
        'properties': {
            'foo': {'type': 'integer'},
            'data': {'type': 'array', 'items': {'$ref': '#/components/schemas/X'}}
        }
    }


@mock.patch('drf_spectacular.settings.spectacular_settings.COMPONENT_SPLIT_REQUEST', True)
def test_serializer_envelope_through_extension(no_warnings):
    class EnvelopeMixin:
        pass

    # actual enveloping not implemented. This could be done internally with
    # to_representation or externally with a custom Renderer
    class XSerializer(EnvelopeMixin, serializers.ModelSerializer):
        name = serializers.CharField()

        class Meta:
            model = SimpleModel
            fields = '__all__'
            envelope = 'foo'  # some arbitrary addition to Meta for example

    class EnvelopeFix(OpenApiSerializerExtension):
        target_class = EnvelopeMixin
        match_subclasses = True

        def get_name(self, auto_schema: 'AutoSchema', direction: Direction):
            if direction == 'request':
                return None
            else:
                return f"Enveloped{self.target.__class__.__name__}"

        def map_serializer(self, auto_schema: 'AutoSchema', direction: Direction):
            if direction == 'request':
                return auto_schema._map_serializer(self.target, direction, bypass_extensions=True)
            else:
                component = auto_schema.resolve_serializer(self.target, direction, bypass_extensions=True)
                if not component:
                    return {}
                return build_object_type(
                    properties={self.target.Meta.envelope: component.ref}
                )

    class XViewset(viewsets.ModelViewSet):
        serializer_class = XSerializer
        queryset = SimpleModel.objects.none()

    schema = generate_schema('/x', XViewset)
    assert 'X' in schema['components']['schemas']
    assert 'EnvelopedX' in schema['components']['schemas']
    assert 'XRequest' in schema['components']['schemas']
    assert 'PatchedXRequest' in schema['components']['schemas']
